import os
from copy import deepcopy
from time import perf_counter

import numcodecs
import numpy as np
import pandas as pd
import pytest
import zarr
from pydantic import ValidationError

from polaris.dataset import DatasetV2, Subset
from polaris.dataset._dataset_v2 import _INDEX_ARRAY_KEY
from polaris.dataset._factory import DatasetFactory
from polaris.dataset.converters._pdb import PDBConverter
from polaris.dataset.zarr._manifest import generate_zarr_manifest


def test_dataset_v2_get_columns(test_dataset_v2):
    assert set(test_dataset_v2.columns) == {"A", "B"}


def test_dataset_v2_get_rows(test_dataset_v2):
    assert set(test_dataset_v2.rows) == set(range(100))


def test_dataset_v2_get_data(test_dataset_v2, zarr_archive):
    root = zarr.open(zarr_archive, "r")
    indices = np.random.randint(0, len(test_dataset_v2), 5)
    for idx in indices:
        assert np.array_equal(test_dataset_v2.get_data(row=idx, col="A"), root["A"][idx])
        assert np.array_equal(test_dataset_v2.get_data(row=idx, col="B"), root["B"][idx])


def test_dataset_v2_with_subset(test_dataset_v2, zarr_archive):
    root = zarr.open(zarr_archive, "r")
    indices = np.random.randint(0, len(test_dataset_v2), 5)
    subset = Subset(test_dataset_v2, indices, "A", "B")
    for i, (x, y) in enumerate(subset):
        idx = indices[i]
        assert np.array_equal(x, root["A"][idx])
        assert np.array_equal(y, root["B"][idx])


def test_dataset_v2_load_to_memory(test_dataset_v2):
    subset = Subset(
        dataset=test_dataset_v2,
        indices=range(100),
        input_cols=["A"],
        target_cols=["B"],
    )

    t1 = perf_counter()
    for x in subset:
        pass
    d1 = perf_counter() - t1

    test_dataset_v2.load_to_memory()

    t2 = perf_counter()
    for x in subset:
        pass
    d2 = perf_counter() - t2

    assert d2 < d1


def test_dataset_v2_serialization(test_dataset_v2, tmp_path):
    save_dir = str(tmp_path / "save_dir")
    path = test_dataset_v2.to_json(save_dir)
    new_dataset = DatasetV2.from_json(path)
    for i in range(5):
        assert np.array_equal(new_dataset.get_data(i, "A"), test_dataset_v2.get_data(i, "A"))
        assert np.array_equal(new_dataset.get_data(i, "B"), test_dataset_v2.get_data(i, "B"))


def test_dataset_v2_caching(test_dataset_v2, tmp_path):
    cache_dir = str(tmp_path / "cache")
    test_dataset_v2._cache_dir = cache_dir
    test_dataset_v2.cache()
    assert str(test_dataset_v2.zarr_root_path).startswith(cache_dir)


def test_dataset_v1_v2_compatibility(test_dataset, tmp_path):
    # A DataFrame is ultimately a collection of labeled numpy arrays
    # We can thus also saved these same arrays to a Zarr archive
    df = test_dataset.table

    path = str(tmp_path / "data" / "v1v2.zarr")

    root = zarr.open(path, "w")
    root.array("smiles", data=df["smiles"].values, dtype=object, object_codec=numcodecs.VLenUTF8())
    root.array("iupac", data=df["iupac"].values, dtype=object, object_codec=numcodecs.VLenUTF8())
    for col in set(df.columns) - {"smiles", "iupac"}:
        root.array(col, data=df[col].values)
    zarr.consolidate_metadata(path)

    kwargs = test_dataset.model_dump(exclude=["table", "zarr_root_path"])
    dataset = DatasetV2(**kwargs, zarr_root_path=path)

    subset_1 = Subset(dataset=test_dataset, indices=range(5), input_cols=["smiles"], target_cols=["calc"])
    subset_2 = Subset(dataset=dataset, indices=range(5), input_cols=["smiles"], target_cols=["calc"])

    for idx in range(5):
        x1, y1 = subset_1[idx]
        x2, y2 = subset_2[idx]
        assert x1 == x2
        assert y1 == y2


def test_dataset_v2_with_pdbs(pdb_paths, tmp_path):
    # The PDB example is interesting because it creates a more complex Zarr archive
    # that includes subgroups
    zarr_root_path = str(tmp_path / "pdbs.zarr")
    factory = DatasetFactory(zarr_root_path)

    # Build a V1 dataset
    converter = PDBConverter()
    factory.register_converter("pdb", converter)
    factory.add_from_files(pdb_paths, axis=0)
    dataset_v1 = factory.build()

    # Build a V2 dataset based on the V1 dataset

    # Add the magic index column to the Zarr subgroup
    root = zarr.open(zarr_root_path, "a")
    ordered_keys = [v.split("/")[-1] for v in dataset_v1.table["pdb"].values]
    root["pdb"].array(_INDEX_ARRAY_KEY, data=ordered_keys, dtype=object, object_codec=numcodecs.VLenUTF8())
    zarr.consolidate_metadata(zarr_root_path)

    # Update annotations to no longer have pointer columns
    annotations = deepcopy(dataset_v1.annotations)
    for anno in annotations.values():
        anno.is_pointer = False

    # Create the dataset
    dataset_v2 = DatasetV2(
        zarr_root_path=zarr_root_path,
        annotations=annotations,
        default_adapters=dataset_v1.default_adapters,
    )

    assert len(dataset_v1) == len(dataset_v2)
    for idx in range(len(dataset_v1)):
        pdb_1 = dataset_v1.get_data(idx, "pdb")
        pdb_2 = dataset_v2.get_data(idx, "pdb")
        assert pdb_1 == pdb_2


def test_dataset_v2_indexing(zarr_archive):
    # Create a subgroup with 100 arrays
    root = zarr.open(zarr_archive, "a")
    subgroup = root.create_group("X")
    for i in range(100):
        subgroup.array(f"{i}", data=np.array([i] * 100))

    # Index it in reverse (element 0 is the last element in the array)
    indices = [f"{idx}" for idx in range(100)][::-1]
    subgroup.array(_INDEX_ARRAY_KEY, data=indices, dtype=object, object_codec=numcodecs.VLenUTF8())
    zarr.consolidate_metadata(zarr_archive)

    # Create the dataset
    dataset = DatasetV2(zarr_root_path=zarr_archive)

    # Check that the dataset is indexed correctly
    for idx in range(100):
        assert np.array_equal(dataset.get_data(row=idx, col="X"), np.array([99 - idx] * 100))


def test_dataset_v2_validation_index_array(zarr_archive):
    root = zarr.open(zarr_archive, "a")

    # Create subgroup that lacks the index array
    subgroup = root.create_group("X")
    zarr.consolidate_metadata(zarr_archive)

    with pytest.raises(ValidationError, match="does not have an index array"):
        DatasetV2(zarr_root_path=zarr_archive)

    indices = [f"{idx}" for idx in range(100)]
    indices[-1] = "invalid"

    # Create index array that doesn't match group length (zero arrays in group, but 100 indices)
    subgroup.array(_INDEX_ARRAY_KEY, data=indices, dtype=object, object_codec=numcodecs.VLenUTF8())
    zarr.consolidate_metadata(zarr_archive)

    with pytest.raises(ValidationError, match="Length of index array"):
        DatasetV2(zarr_root_path=zarr_archive)

    for i in range(100):
        subgroup.array(f"{i}", data=np.random.random(100))
    zarr.consolidate_metadata(zarr_archive)

    # Create index array that has invalid keys (last keys = 'invalid' rather than '99')
    with pytest.raises(ValidationError, match="Keys of index array"):
        DatasetV2(zarr_root_path=zarr_archive)


def test_dataset_v2_validation_consistent_lengths(zarr_archive):
    root = zarr.open(zarr_archive, "a")

    # Change the length of one of the arrays
    root["A"].append(np.random.random((1, 2048)))
    zarr.consolidate_metadata(zarr_archive)

    # Subgroup has a false number of indices
    with pytest.raises(ValidationError, match="should have the same length"):
        DatasetV2(zarr_root_path=zarr_archive)

    # Make the length of the two arrays equal again
    # shouldn't crash
    root["B"].append(np.random.random((1, 2048)))
    zarr.consolidate_metadata(zarr_archive)
    DatasetV2(zarr_root_path=zarr_archive)

    # Create subgroup with inconsistent length
    subgroup = root.create_group("X")
    for i in range(100):
        subgroup.array(f"{i}", data=np.random.random(100))
    indices = [f"{idx}" for idx in range(100)]
    subgroup.array(_INDEX_ARRAY_KEY, data=indices, dtype=object, object_codec=numcodecs.VLenUTF8())
    zarr.consolidate_metadata(zarr_archive)

    # Subgroup has a false number of indices
    with pytest.raises(ValidationError, match="should have the same length"):
        DatasetV2(zarr_root_path=zarr_archive)


def test_zarr_manifest(test_dataset_v2):
    # Assert the manifest Parquet is created
    assert test_dataset_v2.zarr_manifest_path is not None
    assert os.path.isfile(test_dataset_v2.zarr_manifest_path)

    # Assert the manifest contains 204 rows (the number "204" is chosen because
    # the Zarr archive defined in `conftest.py` contains 204 unique files)
    df = pd.read_parquet(test_dataset_v2.zarr_manifest_path)
    assert len(df) == 204

    # Assert the manifest hash is calculated
    assert test_dataset_v2.zarr_manifest_md5sum is not None

    # Add array to Zarr archive to change the number of chunks in the dataset
    root = zarr.open(test_dataset_v2.zarr_root_path, "a")
    root.array("C", data=np.random.random((100, 2048)), chunks=(1, None))

    generate_zarr_manifest(test_dataset_v2.zarr_root_path, test_dataset_v2._cache_dir)

    # Get the length of the updated manifest file
    post_change_manifest_length = len(pd.read_parquet(test_dataset_v2.zarr_manifest_path))

    # Ensure Zarr manifest has an additional 100 chunks + 1 array metadata file
    assert post_change_manifest_length == 305


def test_dataset_v2__get_item__(test_dataset_v2, zarr_archive):
    """Test the __getitem__() interface for the dataset V2."""

    # Ground truth
    root = zarr.open(zarr_archive)

    # Get a specific cell
    assert np.array_equal(test_dataset_v2[0, "A"], root["A"][0, :])

    # Get a specific row
    def _check_row_equality(d1, d2):
        assert len(d1) == len(d2)
        for k in d1:
            assert np.array_equal(d1[k], d2[k])

    _check_row_equality(test_dataset_v2[0], {"A": root["A"][0, :], "B": root["B"][0, :]})
    _check_row_equality(test_dataset_v2[10], {"A": root["A"][10, :], "B": root["B"][10, :]})



import certifi


class InvalidDatasetError(ValueError):
    pass


class InvalidBenchmarkError(ValueError):
    pass


class InvalidCompetitionError(ValueError):
    pass


class InvalidResultError(ValueError):
    pass


class TestAccessError(Exception):
    # Prevent pytest to collect this as a test
    __test__ = False

    pass


class PolarisChecksumError(ValueError):
    pass


class InvalidZarrChecksum(Exception):
    pass


class InvalidZarrCodec(Exception):
    """Raised when an expected codec is not registered."""

    def __init__(self, codec_id: str):
        self.codec_id = codec_id
        super().__init__(
            f"This Zarr archive requires the {self.codec_id} codec. "
            "Install all optional codecs with 'pip install polaris-lib[codecs]'."
        )


class PolarisHubError(Exception):
    BOLD = "\033[1m"
    YELLOW = "\033[93m"
    _END_CODE = "\033[0m"

    def __init__(self, message: str = "", response_text: str = ""):
        parts = filter(
            bool,
            [
                f"{self.BOLD}The request to the Polaris Hub has failed.{self._END_CODE}",
                f"{self.YELLOW}{message}{self._END_CODE}" if message else "",
                f"----------------------\nError reported was:\n{response_text}" if response_text else "",
            ],
        )

        super().__init__("\n".join(parts))


class PolarisUnauthorizedError(PolarisHubError):
    def __init__(self, response_text: str = ""):
        message = (
            "You are not logged in to Polaris or your login has expired. "
            "You can use the Polaris CLI to easily authenticate yourself again with `polaris login --overwrite`."
        )
        super().__init__(message, response_text)


class PolarisCreateArtifactError(PolarisHubError):
    def __init__(self, response_text: str = ""):
        message = (
            "Note: If you can confirm that you are authorized to perform this action, "
            "please call 'polaris login --overwrite' and try again. If the issue persists, please reach out to the Polaris team for support."
        )
        super().__init__(message, response_text)


class PolarisRetrieveArtifactError(PolarisHubError):
    def __init__(self, response_text: str = ""):
        message = (
            "Note: If this artifact exists and you can confirm that you are authorized to retrieve it, "
            "please call 'polaris login --overwrite' and try again. If the issue persists, please reach out to the Polaris team for support."
        )
        super().__init__(message, response_text)


class PolarisSSLError(PolarisHubError):
    def __init__(self, response_text: str = ""):
        message = (
            "We could not verify the SSL certificate. "
            f"Please ensure the installed version ({certifi.__version__}) of the `certifi` package is the latest. "
            "If you require the usage of a custom CA bundle, you can set the POLARIS_CA_BUNDLE "
            "environment variable to the path of your CA bundle. For debugging, you can temporarily disable "
            "SSL verification by setting the POLARIS_CA_BUNDLE environment variable to `false`."
        )
        super().__init__(message, response_text)



import json
import logging
import re
from os import PathLike
from pathlib import Path
from typing import Any, ClassVar, Iterable, Literal

import fsspec
import numpy as np
import zarr
from pydantic import PrivateAttr, computed_field, model_validator
from typing_extensions import Self

from polaris.dataset._adapters import Adapter
from polaris.dataset._base import BaseDataset
from polaris.dataset.zarr._manifest import calculate_file_md5, generate_zarr_manifest
from polaris.utils.errors import InvalidDatasetError
from polaris.utils.types import AccessType, ChecksumStrategy, HubOwner, ZarrConflictResolution

logger = logging.getLogger(__name__)

_INDEX_ARRAY_KEY = "__index__"


class DatasetV2(BaseDataset):
    """Second version of a Polaris Dataset.

    This version gets rid of the DataFrame and stores all data in a Zarr archive.

    V1 stored all datapoints in a Pandas DataFrame. Because a DataFrame is always loaded to memory,
    this was a bottleneck when the number of data points grew large. Even with the pointer columns, you still
    need to load all pointers into memory. V2 therefore switches to a Zarr-only format.

    Attributes:
        zarr_root_path: Required path to a Zarr archive.

    For additional metadata attributes, see the base classes.

    Raises:
        InvalidDatasetError: If the dataset does not conform to the Pydantic data-model specification.
    """

    _artifact_type = "dataset"

    _version: ClassVar[Literal[2]] = 2
    _zarr_manifest_path: str | None = PrivateAttr(None)
    _zarr_manifest_md5sum: str | None = PrivateAttr(None)

    # Redefine this to make it a required field
    zarr_root_path: str

    @model_validator(mode="after")
    def _validate_v2_dataset_model(self) -> Self:
        """Verifies some dependencies between properties"""

        # Since the keys for subgroups are not ordered, we have no easy way to index these groups.
        # Any subgroup should therefore have a special array that defines the index for that group.
        for group in self.zarr_root.group_keys():
            if _INDEX_ARRAY_KEY not in self.zarr_root[group].array_keys():
                raise InvalidDatasetError(f"Group {group} does not have an index array.")

            index_arr = self.zarr_root[group][_INDEX_ARRAY_KEY]
            if len(index_arr) != len(self.zarr_root[group]) - 1:
                raise InvalidDatasetError(
                    f"Length of index array for group {group} does not match the size of the group."
                )
            if any(x not in self.zarr_root[group] for x in index_arr):
                raise InvalidDatasetError(
                    f"Keys of index array for group {group} does not match the group members."
                )

        # Check the structure of the Zarr archive
        # All arrays or groups in the root should have the same length.
        lengths = {len(self.zarr_root[k]) for k in self.zarr_root.array_keys()}
        lengths.update({len(self.zarr_root[k][_INDEX_ARRAY_KEY]) for k in self.zarr_root.group_keys()})
        if len(lengths) > 1:
            raise InvalidDatasetError(
                f"All arrays or groups in the root should have the same length, found the following lengths: {lengths}"
            )

        return self

    @property
    def n_rows(self) -> int:
        """Return all row indices for the dataset"""
        example = self.zarr_root[self.columns[0]]
        match example:
            case zarr.Group():
                return len(example[_INDEX_ARRAY_KEY])
            case _:
                return len(example)

    @property
    def rows(self) -> Iterable[int]:
        """
        Return all row indices for the dataset
        This feature is added for completeness' sake, but does not provide any performance benefits.
        """
        return range(self.n_rows)

    @property
    def columns(self) -> list[str]:
        """Return all columns for the dataset"""
        return list(self.zarr_root.keys())

    @property
    def dtypes(self) -> dict[str, np.dtype]:
        """Return the dtype for each of the columns for the dataset"""
        dtypes = {}
        for arr in self.zarr_root.array_keys():
            dtypes[arr] = self.zarr_root[arr].dtype
        for group in self.zarr_root.group_keys():
            dtypes[group] = np.dtype(object)
        return dtypes

    def load_zarr_root_from_hub(self):
        """
        Loads a Zarr archive from the Hub.
        """
        from polaris.hub.client import PolarisHubClient
        from polaris.hub.storage import StorageSession

        with PolarisHubClient() as client:
            with StorageSession(client, "read", self.urn) as storage:
                return zarr.open_consolidated(store=storage.store("root"))

    @property
    def zarr_manifest_path(self) -> str:
        if self._zarr_manifest_path is None:
            zarr_manifest_path = generate_zarr_manifest(self.zarr_root_path, self._cache_dir)
            self._zarr_manifest_path = zarr_manifest_path

        return self._zarr_manifest_path

    @computed_field
    @property
    def zarr_manifest_md5sum(self) -> str:
        """
        Lazily compute the checksum once needed.

        The checksum of the DatasetV2 is computed from the Zarr Manifest and is _not_ deterministic.
        """
        if not self.has_zarr_manifest_md5sum:
            logger.info("Computing the checksum. This can be slow for large datasets.")
            self.zarr_manifest_md5sum = calculate_file_md5(self.zarr_manifest_path)
        return self._zarr_manifest_md5sum

    @zarr_manifest_md5sum.setter
    def zarr_manifest_md5sum(self, value: str):
        """Set the checksum."""
        if not re.fullmatch(r"^[a-f0-9]{32}$", value):
            raise ValueError("The checksum should be the 32-character hexdigest of a 128 bit MD5 hash.")
        self._zarr_manifest_md5sum = value

    @property
    def has_zarr_manifest_md5sum(self) -> bool:
        """Whether the md5sum for this dataset's zarr manifest has been computed and stored."""
        return self._zarr_manifest_md5sum is not None

    def get_data(self, row: int, col: str, adapters: dict[str, Adapter] | None = None) -> np.ndarray | Any:
        """Indexes the Zarr archive.

        Args:
            row: The index of the data to fetch.
            col: The label of a direct child of the Zarr root.
            adapters: The adapters to apply to the data before returning it.
                If None, will use the default adapters specified for the dataset.

        Returns:
            A numpy array with the data at the specified indices. If the column is a pointer column,
                the content of the referenced file is loaded to memory.
        """
        # Fetch adapters for dataset and a given column
        # Partially override if the adapters parameter is specified.
        adapters = {**self.default_adapters, **(adapters or {})}
        adapter = adapters.get(col)

        # Get the data
        group_or_array = self.zarr_data[col]

        # If it is a group, there is no deterministic order for the child keys.
        # We therefore use a special array that defines the index.
        # If loaded to memory, the group is represented by a dictionary.
        if isinstance(group_or_array, zarr.Group) or isinstance(group_or_array, dict):
            # Indices in a group should always be strings
            row = str(group_or_array[_INDEX_ARRAY_KEY][row])
        arr = group_or_array[row]

        # Adapt the input to the specified format
        if adapter is not None:
            arr = adapter(arr)

        return arr

    def upload_to_hub(
        self,
        access: AccessType = "private",
        owner: HubOwner | str | None = None,
        parent_artifact_id: str | None = None,
    ):
        """
        Uploads the dataset to the Polaris Hub.
        """

        from polaris.hub.client import PolarisHubClient

        with PolarisHubClient() as client:
            client.upload_dataset(self, owner=owner, access=access, parent_artifact_id=parent_artifact_id)

    @classmethod
    def from_json(cls, path: str):
        """
        Loads a dataset from a JSON file.

        Args:
            path: The path to the JSON file to load the dataset from.
        """
        with fsspec.open(path, "r") as f:
            data = json.load(f)
        return cls.model_validate(data)

    def to_json(
        self,
        destination: str | Path,
        if_exists: ZarrConflictResolution = "replace",
    ) -> str:
        """
        Save the dataset to a destination directory as a JSON file.

        Args:
            destination: The _directory_ to save the associated data to.
            if_exists: Action for handling existing files in the Zarr archive. Options are 'raise' to throw
                an error, 'replace' to overwrite, or 'skip' to proceed without altering the existing files.

        Returns:
            The path to the JSON file.
        """
        destination = Path(destination)
        destination.mkdir(exist_ok=True, parents=True)

        # Make a copy to cache the data, and then serialize that
        copy = self.model_copy()
        copy.cache(destination, if_exists=if_exists)

        serialized = copy.model_dump_json(
            exclude_none=True, exclude={"zarr_manifest_path", "zarr_manifest_md5sum"}
        )

        destination_json = destination / f"{copy.slug}.json"
        destination_json.write_text(serialized)
        return str(destination_json)

    def cache(
        self, destination: str | PathLike | None = None, if_exists: ZarrConflictResolution = "replace"
    ) -> str:
        """
        Caches the dataset by downloading the Zarr archive to a local directory.

        Args:
            destination: The directory to cache the data to. If None, will use the default cache directory.
            if_exists: Action for handling existing files at the destination. Options are 'raise' to throw
                an error, 'replace' to overwrite, or 'skip' to proceed without altering the existing files.

        Returns:
            The path to the directory where data has been cached to.
        """
        if not destination:
            destination = self._cache_dir

        destination = Path(destination)
        destination.mkdir(exist_ok=True, parents=True)

        self._cache_zarr(destination, if_exists)

        return str(destination)

    def _repr_dict_(self) -> dict:
        """Utility function for pretty-printing to the command line and jupyter notebooks"""
        repr_dict = self.model_dump(exclude={"zarr_manifest_path", "zarr_manifest_md5sum"})
        return repr_dict

    def should_verify_checksum(self, strategy: ChecksumStrategy) -> bool:
        """
        Determines whether to verify the checksum of the dataset based on the strategy.
        """
        return False



import os
from hashlib import md5
from pathlib import Path

from pyarrow import Table, schema, string
from pyarrow.parquet import write_table

# PyArrow table schema for the V2 Zarr manifest file
ZARR_MANIFEST_SCHEMA = schema([("path", string()), ("md5_checksum", string())])

ROW_GROUP_SIZE = 128 * 1024 * 1024  # 128 MB


def generate_zarr_manifest(zarr_root_path: str, output_dir: str) -> str:
    """
    Entry point function which triggers the creation of a Zarr manifest for a V2 dataset.

    Parameters:
        zarr_root_path: The path to the root of a Zarr archive
        output_dir: The path to the directory which will hold the generated manifest
    """
    zarr_manifest_path = f"{output_dir}/zarr_manifest.parquet"

    entries = manifest_entries(zarr_root_path, zarr_root_path)
    manifest = Table.from_pylist(mapping=entries, schema=ZARR_MANIFEST_SCHEMA)
    write_table(manifest, zarr_manifest_path, row_group_size=ROW_GROUP_SIZE)

    return zarr_manifest_path


def manifest_entries(dir_path: str, root_path: str) -> list[dict[str, str]]:
    """
    Recursive function that traverses a directory, returning entries consisting of every file's path and MD5 hash

    Parameters:
        dir_path: The path to the current directory being traversed
        root_path: The root path from which to compute a relative path
    """
    entries = []
    with os.scandir(dir_path) as it:
        for entry in it:
            if entry.is_file():
                entries.append(
                    {
                        "path": str(Path(entry.path).relative_to(root_path)),
                        "md5_checksum": calculate_file_md5(entry.path),
                    }
                )
            elif entry.is_dir():
                entries.extend(manifest_entries(entry.path, root_path))

    return entries


def calculate_file_md5(file_path: str) -> str:
    """Calculates the md5 hash for a file at a given path"""

    md5_hash = md5()
    with open(file_path, "rb") as file:
        #
        # Read the file in chunks to avoid using too much memory for large files
        for chunk in iter(lambda: file.read(4096), b""):
            md5_hash.update(chunk)

    # Return the hex representation of the digest
    return md5_hash.hexdigest()



import logging

from rich.logging import RichHandler

from ._version import __version__
from .loader import load_benchmark, load_competition, load_dataset, load_model

__all__ = ["load_dataset", "load_benchmark", "load_competition", "load_model", "__version__"]

# Polaris specific logger
logger = logging.getLogger(__name__)

# Only add handler if the logger has not already been configured externally
if not logger.handlers:
    handler = RichHandler(rich_tracebacks=True)
    handler.setFormatter(logging.Formatter("%(message)s", datefmt="[%Y-%m-%d %X]"))
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)



from polaris.evaluate._metadata import ResultsMetadataV1, ResultsMetadataV2
from polaris.evaluate._metadata import ResultsMetadataV1 as ResultsMetadata
from polaris.evaluate._metric import Metric, MetricInfo
from polaris.evaluate._predictions import BenchmarkPredictions, CompetitionPredictions
from polaris.evaluate._results import (
    BenchmarkResultsV1 as BenchmarkResults,
    BenchmarkResultsV1,
    BenchmarkResultsV2,
    CompetitionResults,
    EvaluationResultV1 as EvaluationResult,
    EvaluationResultV1,
    EvaluationResultV2,
)
from polaris.evaluate.utils import evaluate_benchmark

__all__ = [
    "ResultsMetadata",
    "ResultsMetadataV1",
    "ResultsMetadataV2",
    "Metric",
    "MetricInfo",
    "EvaluationResult",
    "EvaluationResultV1",
    "EvaluationResultV2",
    "BenchmarkResults",
    "BenchmarkResultsV1",
    "BenchmarkResultsV2",
    "CompetitionResults",
    "evaluate_benchmark",
    "CompetitionPredictions",
    "BenchmarkPredictions",
]



import numpy as np
import pytest

from polaris.evaluate import BenchmarkPredictions


def assert_deep_equal(result, expected):
    assert isinstance(result, type(expected)), f"Types differ: {type(result)} != {type(expected)}"

    if isinstance(expected, dict):
        assert result.keys() == expected.keys()
        for key in expected:
            assert_deep_equal(result[key], expected[key])
    elif isinstance(expected, np.ndarray):
        assert np.array_equal(result, expected)
    else:
        assert result == expected


def test_benchmark_predictions_normalization():
    # Single task, single test set
    assert_deep_equal(
        {"test": {"col1": np.array([1, 2, 3])}},
        BenchmarkPredictions(
            predictions=[1, 2, 3],
            target_labels=["col1"],
            test_set_labels=["test"],
            test_set_sizes={"test": 3},
        ).predictions,
    )
    assert_deep_equal(
        {"test": {"col1": np.array([1, 2, 3])}},
        BenchmarkPredictions(
            predictions={"test": {"col1": [1, 2, 3]}},
            target_labels=["col1"],
            test_set_labels=["test"],
            test_set_sizes={"test": 3},
        ).predictions,
    )

    # Single task, multiple test sets
    assert_deep_equal(
        {"test": {"col1": np.array([1, 2, 3])}, "test2": {"col1": np.array([4, 5, 6])}},
        BenchmarkPredictions(
            predictions={"test": [1, 2, 3], "test2": [4, 5, 6]},
            target_labels=["col1"],
            test_set_labels=["test", "test2"],
            test_set_sizes={"test": 3, "test2": 3},
        ).predictions,
    )
    assert_deep_equal(
        {"test1": {"col1": np.array([1, 2, 3])}, "test2": {"col1": np.array([4, 5, 6])}},
        BenchmarkPredictions(
            predictions={"test1": {"col1": [1, 2, 3]}, "test2": {"col1": [4, 5, 6]}},
            target_labels=["col1"],
            test_set_labels=["test1", "test2"],
            test_set_sizes={"test1": 3, "test2": 3},
        ).predictions,
    )

    # Multi-task, single test set
    assert_deep_equal(
        {"test": {"col1": np.array([1, 2, 3]), "col2": np.array([4, 5, 6])}},
        BenchmarkPredictions(
            predictions={"col1": [1, 2, 3], "col2": [4, 5, 6]},
            target_labels=["col1", "col2"],
            test_set_labels=["test"],
            test_set_sizes={"test": 3},
        ).predictions,
    )
    assert_deep_equal(
        {"test": {"col1": np.array([1, 2, 3]), "col2": np.array([4, 5, 6])}},
        BenchmarkPredictions(
            predictions={"test": {"col1": [1, 2, 3], "col2": [4, 5, 6]}},
            target_labels=["col1", "col2"],
            test_set_labels=["test"],
            test_set_sizes={"test": 3},
        ).predictions,
    )

    # Multi-task, multiple test sets
    assert_deep_equal(
        {
            "test1": {"col1": np.array([1, 2, 3]), "col2": np.array([4, 5, 6])},
            "test2": {"col1": np.array([7, 8, 9]), "col2": np.array([10, 11, 12])},
        },
        BenchmarkPredictions(
            predictions={
                "test1": {"col1": [1, 2, 3], "col2": [4, 5, 6]},
                "test2": {"col1": [7, 8, 9], "col2": [10, 11, 12]},
            },
            target_labels=["col1", "col2"],
            test_set_labels=["test1", "test2"],
            test_set_sizes={"test1": 3, "test2": 3},
        ).predictions,
    )


def test_benchmark_predictions_incorrect_keys():
    with pytest.raises(ValueError):
        BenchmarkPredictions(
            predictions=[1, 2, 3],
            target_labels=["col1"],
            test_set_labels=["test1", "test2"],
            test_set_sizes={"test1": 3},
        )

    with pytest.raises(ValueError):
        BenchmarkPredictions(
            predictions=[1, 2, 3],
            target_labels=["col1", "col2"],
            test_set_labels=["test1"],
            test_set_sizes={"test1": 3},
        )

    with pytest.raises(ValueError):
        BenchmarkPredictions(
            predictions={"col1": [1, 2, 3]},
            target_labels=["col1"],
            test_set_labels=["test1", "test2"],
            test_set_sizes={"test1": 3, "test2": 3},
        )

    with pytest.raises(ValueError):
        BenchmarkPredictions(
            predictions={"test1": {"col1": [1, 2, 3]}, "test2": {"col1": [4, 5, 6]}},
            target_labels=["col1"],
            test_set_labels=["test1", "test2", "test3"],
            test_set_sizes={"test1": 3, "test2": 3, "test3": 3},
        )


def test_benchmark_predictions_type_checking():
    v1 = {"test": {"col1": ["strings", "also", "valid"]}}
    v2 = BenchmarkPredictions(
        predictions=["strings", "also", "valid"],
        target_labels=["col1"],
        test_set_labels=["test"],
        test_set_sizes={"test": 3},
    ).predictions

    assert list(v1.keys()) == ["test"]
    assert list(v2.keys()) == ["test"]
    assert list(v1["test"].keys()) == ["col1"]
    assert list(v2["test"].keys()) == ["col1"]
    assert isinstance(v2["test"]["col1"], np.ndarray)
    assert np.array_equal(v2["test"]["col1"], np.array(["strings", "also", "valid"]))


def test_invalid_benchmark_predictions_errors():
    with pytest.raises(ValueError):
        BenchmarkPredictions(
            predictions={"test": {"col1": [1, 2, 3]}, "test2": [4, 5, 6]},
            target_cols=["col1", "col2"],
            test_set_labels=["test", "test2"],
            test_set_sizes={"test": 3, "test2": 3},
        )

    with pytest.raises(ValueError):
        BenchmarkPredictions(
            predictions={"test": {"col1": "not an array or list"}},
            target_cols=["col1"],
            test_set_labels=["test"],
            test_set_sizes={"test": 1},
        )

    with pytest.raises(ValueError):
        BenchmarkPredictions(
            predictions={"test": {"wrong column name": [1, 2, 3]}},
            target_cols=["col1"],
            test_set_labels=["test"],
            test_set_sizes={"test": 3},
        )

    # You should either fully or minimally specify the predictions.
    # We don't allow in-between results.
    with pytest.raises(ValueError):
        BenchmarkPredictions(
            predictions={"col1": [1, 2, 3]},
            target_cols=["col1"],
            test_set_labels=["test"],
            test_set_sizes={"test": 3},
        )

    with pytest.raises(ValueError):
        BenchmarkPredictions(
            predictions={"test": [1, 2, 3]},
            target_cols=["col1"],
            test_set_labels=["test"],
            test_set_sizes={"test": 3},
        )

    # You shouldn't specify more keys than expected.
    with pytest.raises(ValueError):
        BenchmarkPredictions(
            predictions={"test": {"col1": [1, 2, 3], "col2": [4, 5, 6]}},
            target_cols=["col1"],
            test_set_labels=["test"],
            test_set_sizes={"test": 3},
        )
    with pytest.raises(ValueError):
        BenchmarkPredictions(
            predictions={"test": {"col1": [1, 2, 3]}, "test2": {"col1": [1, 2, 3]}},
            target_cols=["col1"],
            test_set_labels=["test"],
            test_set_sizes={"test": 3, "test2": 3},
        )

    # Incorrect size
    with pytest.raises(ValueError):
        BenchmarkPredictions(
            predictions={"test": {"col": [1, 2, 3, 4]}},
            target_cols=["col1"],
            test_set_labels=["test"],
            test_set_sizes={"test": 3},
        )

    # Invalid test_set_sizes
    with pytest.raises(ValueError):
        BenchmarkPredictions(
            predictions={"test": {"col": [1, 2, 3]}},
            target_cols=["col1"],
            test_set_labels=["test"],
            test_set_sizes={"test1": 3},
        )


def test_benchmark_predictions_serialization():
    predictions = BenchmarkPredictions(
        predictions=[1, 2, 3],
        target_labels=["col1"],
        test_set_labels=["test"],
        test_set_sizes={"test": 3},
    )
    serialized = predictions.model_dump()
    assert serialized["predictions"] == {"test": {"col1": [1, 2, 3]}}
    assert serialized["target_labels"] == ["col1"]
    assert serialized["test_set_labels"] == ["test"]

    deserialized = BenchmarkPredictions(**serialized)
    assert set(deserialized.predictions.keys()) == {"test"}
    assert set(deserialized.predictions["test"].keys()) == {"col1"}
    assert np.array_equal(deserialized.predictions["test"]["col1"], np.array([1, 2, 3]))
    assert deserialized.target_labels == ["col1"]
    assert deserialized.test_set_labels == ["test"]
    assert set(deserialized.test_set_sizes.keys()) == {"test"}
    assert deserialized.test_set_sizes["test"] == 3


def test_benchmark_predictions_size():
    predictions = BenchmarkPredictions(
        predictions={
            "test1": {"col1": [1, 2, 3], "col2": [4, 5, 6]},
            "test2": {"col1": [7, 8], "col2": [9, 10]},
        },
        target_labels=["col1", "col2"],
        test_set_labels=["test1", "test2"],
        test_set_sizes={"test1": 3, "test2": 2},
    )
    assert len(predictions) == 10
    assert predictions.get_size() == 10
    assert predictions.get_size(test_set_subset=["test1"]) == 6
    assert predictions.get_size(test_set_subset=["test2"]) == 4
    assert predictions.get_size(target_subset=["col1"]) == 5
    assert predictions.get_size(target_subset=["col2"]) == 5
    assert predictions.get_size(test_set_subset=["test1"], target_subset=["col2"]) == 3



from typing import Any

import numpy as np

from polaris.utils.types import (
    ListOrArrayType,
    SlugCompatibleStringType,
    SlugStringType,
    ArtifactUrn,
    HubOwner,
)


def listit(t: Any):
    """
    Converts all tuples in a possibly nested object to lists
    https://stackoverflow.com/questions/1014352/how-do-i-convert-a-nested-tuple-of-tuples-and-lists-to-lists-of-lists-in-python
    """
    return list(map(listit, t)) if isinstance(t, (list, tuple)) else t


def slugify(sluggable: SlugCompatibleStringType) -> SlugStringType:
    """
    Converts a slug-compatible string to a slug.
    """
    return sluggable.lower().replace("_", "-").strip("-")


def convert_lists_to_arrays(predictions: ListOrArrayType | dict) -> np.ndarray | dict:
    """
    Recursively converts all plain Python lists in the predictions object to numpy arrays
    """

    def convert_to_array(v):
        if isinstance(v, np.ndarray):
            return v
        elif isinstance(v, list):
            return np.array(v)
        elif isinstance(v, dict):
            return {k: convert_to_array(v) for k, v in v.items()}

    return convert_to_array(predictions)


def build_urn(artifact_type: str, owner: str | HubOwner, slug: str) -> ArtifactUrn:
    return f"urn:polaris:{artifact_type}:{owner}:{slug}"



from polaris.benchmark._base import BenchmarkSpecification, BenchmarkV1Specification
from polaris.benchmark._benchmark_v2 import BenchmarkV2Specification
from polaris.benchmark._definitions import MultiTaskBenchmarkSpecification, SingleTaskBenchmarkSpecification

__all__ = [
    "BenchmarkSpecification",
    "BenchmarkV1Specification",
    "BenchmarkV2Specification",
    "SingleTaskBenchmarkSpecification",
    "MultiTaskBenchmarkSpecification",
]



import datamol as dm
import zarr

from polaris.dataset.zarr.codecs import RDKitMolCodec


def test_rdkit_mol_codec():
    mol = dm.to_mol("C1=CC=CC=C1")

    arr = zarr.empty(shape=10, chunks=2, dtype=object, object_codec=RDKitMolCodec())

    arr[0] = mol
    arr[1] = mol
    arr[2] = mol

    assert dm.same_mol(arr[0], mol)
    assert dm.same_mol(arr[1], mol)
    assert dm.same_mol(arr[2], mol)



class FormattingMixin:
    """Mixin class for formatting strings to be output in the console"""

    BOLD = "\033[1m"
    YELLOW = "\033[93m"
    _END_CODE = "\033[0m"

    def format(self, text: str, codes: str | list[str]):
        if not isinstance(codes, list):
            codes = [codes]

        return "".join(codes) + text + self._END_CODE



import json
import logging
from typing import Annotated, ClassVar, Literal, Optional

import fsspec
from packaging.version import Version
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    computed_field,
    field_serializer,
    field_validator,
)
from pydantic.alias_generators import to_camel
from typing_extensions import Self

import polaris
from polaris.utils.misc import build_urn, slugify
from polaris.utils.types import ArtifactUrn, HubOwner, SlugCompatibleStringType, SlugStringType

logger = logging.getLogger(__name__)


class BaseArtifactModel(BaseModel):
    """
    Base class for all artifacts on the Hub. Specifies metadata that is used by the Hub.

    Info: Optional
        Despite all artifacts basing this class, note that all attributes are optional.
        This ensures the library can be used without the Polaris Hub.
        Only when uploading to the Hub, some of the attributes are required.

    Attributes:
        name: A slug-compatible name for the artifact.
            Together with the owner, this is used by the Hub to uniquely identify the artifact.
        description: A beginner-friendly, short description of the artifact.
        tags: A list of tags to categorize the artifact by. This is used by the Hub to search over artifacts.
        user_attributes: A dict with additional, textual user attributes.
        owner: A slug-compatible name for the owner of the artifact.
            If the artifact comes from the Polaris Hub, this is the associated owner (organization or user).
            Together with the name, this is used by the Hub to uniquely identify the artifact.
        polaris_version: The version of the Polaris library that was used to create the artifact.
    """

    _version: ClassVar[Literal[1]] = 1
    _artifact_type: ClassVar[str]

    model_config = ConfigDict(alias_generator=to_camel, populate_by_name=True, arbitrary_types_allowed=True)

    # Model attributes
    name: SlugCompatibleStringType | None = None
    description: str = ""
    tags: list[str] = Field(default_factory=list)
    user_attributes: dict[str, str] = Field(default_factory=dict)
    owner: HubOwner | None = None
    polaris_version: str = polaris.__version__
    slug: Annotated[Optional[SlugStringType], Field(validate_default=True)] = None

    @field_validator("slug")
    def _validate_slug(cls, val: Optional[str], info) -> SlugStringType | None:
        # A slug may be None when an artifact is created locally
        if val is None:
            if info.data.get("name") is not None:
                return slugify(info.data.get("name"))
        return val

    @computed_field
    @property
    def artifact_id(self) -> str | None:
        if self.owner and self.slug:
            return f"{self.owner}/{self.slug}"
        return None

    @computed_field
    @property
    def urn(self) -> ArtifactUrn | None:
        if self.owner and self.slug:
            return self.urn_for(self.owner, self.slug)
        return None

    @computed_field
    @property
    def version(self) -> int:
        return self._version

    @field_validator("polaris_version")
    @classmethod
    def _validate_version(cls, value: str) -> str:
        if value != "dev":
            # Make sure it is a valid semantic version
            Version(value)

        current_version = polaris.__version__
        if value != current_version:
            logger.info(
                f"The version of Polaris that was used to create the artifact ({value}) is different "
                f"from the currently installed version of Polaris ({current_version})."
            )
        return value

    @field_validator("owner", mode="before")
    @classmethod
    def _validate_owner(cls, value: str | HubOwner | None):
        if isinstance(value, str):
            return HubOwner(slug=value)
        return value

    @field_serializer("owner")
    def _serialize_owner(self, value: HubOwner) -> str | None:
        return value.slug if value else None

    @classmethod
    def from_json(cls, path: str) -> Self:
        """Loads an artifact from a JSON file.

        Args:
            path: Path to a JSON file containing the artifact definition.
        """
        with fsspec.open(path, "r") as f:
            data = json.load(f)
            return cls.model_validate(data)

    def to_json(self, path: str) -> None:
        """Saves an artifact to a JSON file.

        Args:
            path: Path to save the artifact definition as JSON.
        """
        with fsspec.open(path, "w") as f:
            f.write(self.model_dump_json())

    @classmethod
    def urn_for(cls, owner: str | HubOwner, slug: str) -> ArtifactUrn:
        return build_urn(cls._artifact_type, owner, slug)



from itertools import chain

import pytest
from pydantic import ValidationError

from polaris.competition import CompetitionSpecification
from polaris.evaluate import Metric
from polaris.utils.types import TaskType


def test_competition_split_verification(test_competition):
    """Verifies that the split validation works as expected."""

    obj = test_competition
    cls = CompetitionSpecification

    # By using the fixture as a default, we know it doesn't always fail
    default_kwargs = {
        "target_cols": obj.target_cols,
        "input_cols": obj.input_cols,
        "metrics": obj.metrics,
        "main_metric": obj.main_metric,
        "name": obj.name,
        "zarr_root_path": obj.zarr_root_path,
        "readme": obj.readme,
        "start_time": obj.start_time,
        "end_time": obj.end_time,
        "n_test_sets": obj.n_test_sets,
        "n_test_datapoints": obj.n_test_datapoints,
        "n_classes": obj.n_classes,
    }

    train_split = obj.split[0]
    test_split = obj.split[1]

    # One or more empty test partitions
    with pytest.raises(ValidationError):
        cls(split=(train_split,), **default_kwargs)
    with pytest.raises(ValidationError):
        cls(split=(train_split, []), **default_kwargs)
    with pytest.raises(ValidationError):
        cls(split=(train_split, {"test": []}), **default_kwargs)
    # Non-exclusive partitions
    with pytest.raises(ValidationError):
        cls(split=(train_split, test_split["test"] + train_split[:1]), **default_kwargs)
    with pytest.raises(ValidationError):
        cls(split=(train_split, {"test1": test_split, "test2": train_split[:1]}), **default_kwargs)
    # Invalid indices
    with pytest.raises(ValidationError):
        cls(split=(train_split + [len(obj)], test_split), **default_kwargs)
    with pytest.raises(ValidationError):
        cls(split=(train_split + [-1], test_split), **default_kwargs)
    with pytest.raises(ValidationError):
        cls(split=(train_split, test_split["test"] + [len(obj)]), **default_kwargs)
    with pytest.raises(ValidationError):
        cls(split=(train_split, test_split["test"] + [-1]), **default_kwargs)
    # Duplicate indices
    with pytest.raises(ValidationError):
        cls(split=(train_split + train_split[:1], test_split), **default_kwargs)
    with pytest.raises(ValidationError):
        cls(split=(train_split, test_split["test"] + test_split["test"][:1]), **default_kwargs)
    with pytest.raises(ValidationError):
        cls(
            split=(train_split, {"test1": test_split, "test2": test_split["test"] + test_split["test"][:1]}),
            **default_kwargs,
        )

    # It should _not_ fail with duplicate indices across test partitions
    cls(split=(train_split, {"test1": test_split["test"], "test2": test_split["test"]}), **default_kwargs)
    # It should _not_ fail with missing indices
    cls(split=(train_split[:-1], test_split), **default_kwargs)
    # It should _not_ fail with an empty train set
    competition = cls(split=([], test_split), **default_kwargs)
    train, _ = competition.get_train_test_split()
    assert len(train) == 0


@pytest.mark.parametrize("cls", [CompetitionSpecification])
def test_competition_metrics_verification(test_competition, cls):
    """Verifies that the metric validation works as expected."""
    # By using the fixture as a default, we know it doesn't always fail
    base = test_competition

    default_kwargs = {
        "target_cols": base.target_cols,
        "input_cols": base.input_cols,
        "main_metric": base.main_metric,
        "name": base.name,
        "zarr_root_path": base.zarr_root_path,
        "readme": base.readme,
        "start_time": base.start_time,
        "end_time": base.end_time,
        "n_test_sets": base.n_test_sets,
        "n_test_datapoints": base.n_test_datapoints,
        "n_classes": base.n_classes,
    }

    # Invalid metric
    with pytest.raises(ValidationError):
        cls(metrics=["invalid"], **default_kwargs)
    with pytest.raises(ValidationError):
        cls(metrics="invalid", **default_kwargs)
    with pytest.raises(ValidationError):
        metrics_list = list(base.metrics)
        cls(
            metrics=metrics_list + [metrics_list[0]],
            **default_kwargs,
        )


def test_competition_duplicate_metrics(test_competition):
    """Tests that passing duplicate metrics will raise a validation error"""
    m = test_competition.model_dump()

    with pytest.raises(ValidationError, match="The benchmark specifies duplicate metric"):
        m["metrics"] = [
            Metric(label="roc_auc", config={"group_by": "CLASS_expt"}),
            Metric(label="roc_auc", config={"group_by": "CLASS_expt"}),
        ]
        m["main_metric"] = m["metrics"][0]
        CompetitionSpecification(**m)

    with pytest.raises(ValidationError, match="The metrics of a benchmark need to have unique names."):
        m["metrics"][0].config.group_by = "MULTICLASS_calc"
        CompetitionSpecification(**m)

    m["metrics"][0].custom_name = "custom_name"
    CompetitionSpecification(**m)


def test_competition_metric_deserialization(test_competition):
    """Tests that passing metrics as a list of strings or dictionaries works as expected"""
    m = test_competition.model_dump()

    # Should work with strings
    m["metrics"] = ["mean_absolute_error", "accuracy"]
    m["main_metric"] = "accuracy"
    CompetitionSpecification(**m)

    # Should work with dictionaries
    m["metrics"] = [
        {"label": "mean_absolute_error", "config": {"group_by": "CLASS_expt"}},
        {"label": "accuracy"},
    ]
    CompetitionSpecification(**m)


def test_competition_train_test_split(test_competition):
    """Tests that the competition's train/test split can be retrieved through a CompetitionSpecification instance"""

    train, test = test_competition.get_train_test_split()

    train_split = test_competition.split[0]
    test_sets = test_competition.split[1]
    test_split = set(chain.from_iterable(test_sets.values()))

    assert len(train) == len(train_split)
    assert len(test) == len(test_split)


def test_competition_computed_fields(test_competition):
    default_test_set_name = "test"
    assert test_competition.task_type == TaskType.SINGLE_TASK.value
    assert test_competition.test_set_labels == [default_test_set_name]
    assert test_competition.test_set_sizes == {default_test_set_name: 10}


def test_competition_interface(test_competition):
    """Tests that the CompetitionSpecification class doesn't accidentally inherit the evaluate method from the benchmark class"""
    with pytest.raises(AttributeError):
        test_competition.evaluate()



import numpy as np
from scipy import stats
from sklearn.metrics import average_precision_score as sk_average_precision_score
from sklearn.metrics import cohen_kappa_score as sk_cohen_kappa_score


def pearsonr(y_true: np.ndarray, y_pred: np.ndarray):
    """Calculate a pearson r correlation"""
    return stats.pearsonr(y_true, y_pred).statistic


def spearman(y_true: np.ndarray, y_pred: np.ndarray):
    """Calculate a Spearman correlation"""
    return stats.spearmanr(y_true, y_pred).statistic


def absolute_average_fold_error(y_true: np.ndarray, y_pred: np.ndarray) -> float:
    """
    Calculate the Absolute Average Fold Error (AAFE) metric.
    It measures the fold change between predicted values and observed values.
    The implementation is based on [this paper](https://pubs.acs.org/doi/10.1021/acs.chemrestox.3c00305).

    Args:
        y_true: The true target values of shape (n_samples,)
        y_pred: The predicted target values of shape (n_samples,).

    Returns:
        aafe: The Absolute Average Fold Error.
    """
    if len(y_true) != len(y_pred):
        raise ValueError("Length of y_true and y_pred must be the same.")

    if np.any(y_true == 0):
        raise ValueError("`y_true` contains zero which will result `Inf` value.")

    aafe = np.mean(np.abs(y_pred) / np.abs(y_true))

    return aafe


def cohen_kappa_score(y_true, y_pred, **kwargs):
    """Scikit learn cohen_kappa_score wraper with renamed arguments"""
    return sk_cohen_kappa_score(y1=y_true, y2=y_pred, **kwargs)


def average_precision_score(y_true, y_score, **kwargs):
    """Scikit learn average_precision_score wrapper that throws an error if y_true has no positive class"""
    if len(y_true) == 0 or not np.any(y_true):
        raise ValueError("Average precision requires at least a single positive class")
    return sk_average_precision_score(y_true=y_true, y_score=y_score, **kwargs)



from typing import Collection

from pydantic import computed_field, field_validator

from polaris.benchmark._base import BenchmarkV1Specification
from polaris.utils.types import TaskType


class SingleTaskMixin:
    """
    Mixin for single-task benchmarks.
    """

    @field_validator("target_cols", check_fields=False)
    @classmethod
    def validate_target_cols(cls, v: Collection[str]) -> Collection[str]:
        if len(v) != 1:
            raise ValueError("A single-task benchmark should specify exactly one target column.")
        return v

    @computed_field
    @property
    def task_type(self) -> str:
        """Return SINGLE_TASK for single-task benchmarks."""
        return TaskType.SINGLE_TASK.value


class MultiTaskMixin:
    """
    Mixin for multi-task benchmarks.
    """

    @field_validator("target_cols", check_fields=False)
    @classmethod
    def validate_target_cols(cls, v: Collection[str]) -> Collection[str]:
        if len(v) <= 1:
            raise ValueError("A multi-task benchmark should specify at least two target columns.")
        return v

    @computed_field
    @property
    def task_type(self) -> str:
        """
        Return MULTI_TASK for multi-task benchmarks.
        """
        return TaskType.MULTI_TASK.value


class SingleTaskBenchmarkSpecification(SingleTaskMixin, BenchmarkV1Specification):
    """
    Single-task benchmark for the base specification.
    """

    pass


class MultiTaskBenchmarkSpecification(MultiTaskMixin, BenchmarkV1Specification):
    """
    Multitask benchmark for the base specification.
    """

    pass



from datetime import datetime

from pydantic import Field, PrivateAttr, computed_field

from polaris._artifact import BaseArtifactModel
from polaris.utils.dict2html import dict2html
from polaris.utils.types import HttpUrlString, HubUser
from polaris.model import Model


class ResultsMetadataV1(BaseArtifactModel):
    """V1 implementation of evaluation results without model field support

    Attributes:
        github_url: The URL to the code repository that was used to generate these results.
        paper_url: The URL to the paper describing the methodology used to generate these results.
        contributors: The users that are credited for these results.

    For additional metadata attributes, see the base classes.
    """

    # Additional metadata
    github_url: HttpUrlString | None = Field(None, alias="code_url")
    paper_url: HttpUrlString | None = Field(None, alias="report_url")
    contributors: list[HubUser] = Field(default_factory=list)

    # Private attributes
    _created_at: datetime = PrivateAttr(default_factory=datetime.now)

    def _repr_html_(self) -> str:
        """For pretty-printing in Jupyter Notebooks"""
        return dict2html(self.model_dump())

    def __repr__(self):
        return self.model_dump_json(indent=2)


class ResultsMetadataV2(BaseArtifactModel):
    """V2 implementation of evaluation results with model field replacing URLs

    Attributes:
        model: The model that was used to generate these results.
        contributors: The users that are credited for these results.

    For additional metadata attributes, see the base classes.
    """

    # Additional metadata
    model: Model | None = Field(None, exclude=True)
    contributors: list[HubUser] = Field(default_factory=list)

    # Private attributes
    _created_at: datetime = PrivateAttr(default_factory=datetime.now)

    @computed_field
    @property
    def model_artifact_id(self) -> str:
        return self.model.artifact_id

    def _repr_html_(self) -> str:
        return dict2html(self.model_dump())

    def __repr__(self):
        return self.model_dump_json(indent=2)



import pytest
from pydantic import ValidationError
from pyroaring import BitMap

from polaris.benchmark._benchmark_v2 import BenchmarkV2Specification
from polaris.benchmark._split_v2 import IndexSet, SplitV2


@pytest.fixture
def index_set():
    return IndexSet(indices=BitMap([1, 2, 3, 4, 5]))


@pytest.fixture
def valid_split():
    return SplitV2(training=IndexSet(indices=BitMap([0, 1, 2])), test=IndexSet(indices=BitMap([3, 4, 5])))


def test_index_set_creation():
    """Test IndexSet initialization and validation"""
    # Test with sequence
    index_set = IndexSet(indices=[1, 2, 3])
    assert isinstance(index_set.indices, BitMap)
    assert index_set.datapoints == 3

    # Test with BitMap
    bitmap = BitMap([1, 2, 3])
    index_set = IndexSet(indices=bitmap)
    assert isinstance(index_set.indices, BitMap)
    assert index_set.datapoints == 3


def test_index_set_serialization(index_set):
    """
    Test IndexSet serialization and deserialization
    """
    serialized = index_set.serialize()
    deserialized = IndexSet.deserialize(serialized)
    assert deserialized.indices == index_set.indices
    assert deserialized.datapoints == index_set.datapoints
    assert deserialized.md5_checksum == index_set.md5_checksum


def test_split_v2_serialized_initialization(valid_split):
    """
    Test that SplitV2 can be initialized from serialized IndexSets
    """
    test = valid_split.test.serialize()
    training = valid_split.training.serialize()

    split = SplitV2(training=training, test=test)
    assert split.training.indices == valid_split.training.indices
    assert split.test.indices == valid_split.test.indices


def test_split_v2_validation():
    """Test SplitV2 validation rules"""
    # Test empty test set
    with pytest.raises(ValidationError):
        SplitV2(training=IndexSet(indices=BitMap([1, 2])), test=IndexSet(indices=BitMap([])))

    # Test overlapping sets
    with pytest.raises(ValidationError):
        SplitV2(training=IndexSet(indices=BitMap([1, 2])), test=IndexSet(indices=BitMap([2, 3])))

    # Valid split should not raise
    SplitV2(training=IndexSet(indices=BitMap([1, 2])), test=IndexSet(indices=BitMap([3, 4])))


def test_split_v2_properties(valid_split):
    """
    Test the properties of SplitV2 reflect the underlying index sets
    """
    assert valid_split.n_train_datapoints == valid_split.training.datapoints
    assert valid_split.n_test_datapoints == {"test": valid_split.test.datapoints}
    assert valid_split.n_test_sets == 1
    assert valid_split.max_index == 5


def test_benchmark_v2_specification(valid_split, test_dataset_v2, tmp_path):
    """
    Test BenchmarkV2Specification initialization and validation
    """
    # Mock minimal valid configuration
    config = {
        "dataset": test_dataset_v2,
        "split": valid_split,
        "target_cols": ["B"],
        "input_cols": ["A"],
        "metrics": ["accuracy", "r2"],
        "main_metric": "accuracy",
        "n_classes": {"B": 2},
        "name": "test_benchmark",
    }

    # Test valid configuration
    benchmark = BenchmarkV2Specification(**config)
    assert benchmark.n_train_datapoints == 3
    assert benchmark.n_test_datapoints == {"test": 3}
    assert benchmark.test_set_labels == ["test"]

    # Test dict as dataset input
    config = {
        "dataset": test_dataset_v2.model_dump(),
        "split": valid_split,
        "target_cols": ["B"],
        "input_cols": ["A"],
        "metrics": ["accuracy", "r2"],
        "main_metric": "accuracy",
        "n_classes": {"B": 2},
        "name": "test_benchmark",
    }
    BenchmarkV2Specification(**config)


def test_benchmark_v2_invalid_indices(valid_split, test_dataset_v2):
    """
    Test validation of indices against dataset length
    """
    max_index = len(test_dataset_v2)
    config = {
        "dataset": test_dataset_v2,
        "split": SplitV2(
            training=IndexSet(indices=BitMap([0, 1])),
            test=IndexSet(indices=BitMap([max_index, max_index + 1])),  # Invalid indices
        ),
        "target_cols": ["B"],
        "input_cols": ["A"],
        "metrics": ["accuracy", "r2"],
        "main_metric": "accuracy",
        "n_classes": {"B": 2},
        "name": "test_benchmark",
    }

    with pytest.raises(ValidationError):
        BenchmarkV2Specification(**config)


def test_benchmark_v2_n_classes_validation(test_dataset_v2):
    """Test n_classes validation"""
    config = {
        "dataset": test_dataset_v2,
        "split": SplitV2(training=IndexSet(indices=BitMap([0, 1])), test=IndexSet(indices=BitMap([2, 3]))),
        "target_cols": ["B"],
        "input_cols": ["A"],
        "metrics": ["accuracy", "r2"],
        "main_metric": "accuracy",
        "n_classes": {"invalid_target": 2},  # Invalid target column
        "name": "test_benchmark",
    }

    with pytest.raises(ValidationError):
        BenchmarkV2Specification(**config)



from collections import defaultdict
from datetime import datetime
from itertools import chain
from typing import Callable

from pydantic import Field, computed_field, model_validator
from typing_extensions import Self

from polaris.benchmark._split import SplitSpecificationV1Mixin
from polaris.benchmark._task import PredictiveTaskSpecificationMixin
from polaris.dataset import DatasetV2, Subset
from polaris.evaluate import CompetitionPredictions
from polaris.utils.dict2html import dict2html
from polaris.utils.errors import InvalidCompetitionError
from polaris.utils.types import (
    ColumnName,
    HttpUrlString,
    HubOwner,
    HubUser,
    IncomingPredictionsType,
    SlugCompatibleStringType,
)


class CompetitionSpecification(DatasetV2, PredictiveTaskSpecificationMixin, SplitSpecificationV1Mixin):
    """An instance of this class represents a Polaris competition.

    Examples:
        Basic API usage:
        ```python
        import polaris as po

        # Load the benchmark from the Hub
        competition = po.load_competition("dummy-user/dummy-name")

        # Get the train and test data-loaders
        train, test = competition.get_train_test_split()

        # Use the training data to train your model
        # Get the input as an array with 'train.inputs' and 'train.targets'
        # Or simply iterate over the train object.
        for x, y in train:
            ...

        # Work your magic to accurately predict the test set
        prediction_values = np.array([0.0 for x in test])

        # Submit your predictions
        competition.submit_predictions(
            prediction_name="first-prediction",
            prediction_owner="dummy-user",
            report_url="REPORT_URL",
            predictions=prediction_values,
        )
        ```

    Attributes:
        start_time: The time at which the competition starts accepting prediction submissions.
        end_time: The time at which the competition stops accepting prediction submissions.
        n_classes: The number of classes within each target column that defines a classification task.

    For additional metadata attributes, see the base classes.
    """

    _artifact_type = "competition"

    dataset: None = None

    start_time: datetime
    end_time: datetime
    n_classes: dict[ColumnName, int | None] = Field(..., default_factory=lambda: defaultdict(None))

    @model_validator(mode="after")
    def _validate_split_in_dataset(self) -> Self:
        """
        All indices are valid given the dataset. We check the len of `self` here because a
        competition entity includes both the dataset and benchmark in one artifact.
        """
        max_i = len(self)
        if any(i < 0 or i >= max_i for i in chain(self.split[0], *self.split[1].values())):
            raise InvalidCompetitionError("The predefined split contains invalid indices")

        return self

    @model_validator(mode="after")
    def _validate_cols_in_dataset(self) -> Self:
        """
        Verifies that all specified columns are present in the dataset.
        """
        columns = self.target_cols | self.input_cols
        dataset_columns = set(self.columns)
        if not columns.issubset(dataset_columns):
            raise InvalidCompetitionError("Not all target or input columns were found in the dataset.")

        return self

    @model_validator(mode="after")
    def _validate_n_classes(self) -> Self:
        """
        The number of classes for each of the target columns.
        """
        columns = set(self.n_classes.keys())
        if not columns.issubset(self.target_cols):
            raise InvalidCompetitionError("Not all target class members were found in the target columns.")

        return self

    @computed_field
    @property
    def dataset_artifact_id(self) -> str:
        return self.artifact_id

    def _get_subset(self, indices, hide_targets=True, featurization_fn=None) -> Subset:
        """Returns a [`Subset`][polaris.dataset.Subset] using the given indices. Used
        internally to construct the train and test sets."""
        return Subset(
            dataset=self,
            indices=indices,
            input_cols=self.input_cols,
            target_cols=self.target_cols,
            hide_targets=hide_targets,
            featurization_fn=featurization_fn,
        )

    def _get_test_sets(
        self, hide_targets=True, featurization_fn: Callable | None = None
    ) -> dict[str, Subset]:
        """
        Construct the test set(s), given the split in the competition specification. Used
        internally to construct the test set for client use and evaluation.
        """
        test_split = self.split[1]
        return {
            k: self._get_subset(v, hide_targets=hide_targets, featurization_fn=featurization_fn)
            for k, v in test_split.items()
        }

    def get_train_test_split(
        self, featurization_fn: Callable | None = None
    ) -> tuple[Subset, Subset | dict[str, Subset]]:
        """Construct the train and test sets, given the split in the competition specification.

        Returns [`Subset`][polaris.dataset.Subset] objects, which offer several ways of accessing the data
        and can thus easily serve as a basis to build framework-specific (e.g. PyTorch, Tensorflow)
        data-loaders on top of.

        Args:
            featurization_fn: A function to apply to the input data. If a multi-input benchmark, this function
                expects an input in the format specified by the `input_format` parameter.

        Returns:
            A tuple with the train `Subset` and test `Subset` objects.
                If there are multiple test sets, these are returned in a dictionary and each test set has
                an associated name. The targets of the test set can not be accessed.
        """

        train = self._get_subset(self.split[0], hide_targets=False, featurization_fn=featurization_fn)
        test = self._get_test_sets(hide_targets=True, featurization_fn=featurization_fn)

        # For improved UX, we return the object instead of the dictionary if there is only one test set.
        # Internally, however, assume that the test set is always a dictionary simplifies the code.
        if len(test) == 1:
            test = test["test"]
        return train, test

    def submit_predictions(
        self,
        predictions: IncomingPredictionsType,
        prediction_name: SlugCompatibleStringType,
        prediction_owner: str,
        report_url: HttpUrlString,
        contributors: list[HubUser] | None = None,
        github_url: HttpUrlString | None = None,
        description: str = "",
        tags: list[str] | None = None,
        user_attributes: dict[str, str] | None = None,
    ) -> None:
        """
        Convenient wrapper around the
        [`PolarisHubClient.submit_competition_predictions`][polaris.hub.client.PolarisHubClient.submit_competition_predictions] method.
        It handles the creation of a standardized predictions object, which is expected by the Hub, automatically.

        Args:
            prediction_name: The name of the prediction.
            prediction_owner: The slug of the user/organization which owns the prediction.
            predictions: The predictions for each test set defined in the competition.
            report_url: A URL to a report/paper/write-up which describes the methods used to generate the predictions.
            contributors: The users credited with generating these predictions.
            github_url: An optional URL to a code repository containing the code used to generated these predictions.
            description: An optional and short description of the predictions.
            tags: An optional list of tags to categorize the prediction by.
            user_attributes: An optional dict with additional, textual user attributes.
        """
        from polaris.hub.client import PolarisHubClient

        standardized_predictions = CompetitionPredictions(
            name=prediction_name,
            owner=HubOwner(slug=prediction_owner),
            predictions=predictions,
            report_url=report_url,
            contributors=contributors or [],
            github_url=github_url,
            description=description,
            tags=tags or [],
            user_attributes=user_attributes or {},
            target_labels=self.target_cols,
            test_set_labels=self.test_set_labels,
            test_set_sizes=self.test_set_sizes,
        )

        with PolarisHubClient() as client:
            client.submit_competition_predictions(
                competition=self, competition_predictions=standardized_predictions
            )

    def _repr_html_(self):
        """For pretty printing in Jupyter."""
        return dict2html(self.model_dump(exclude={"zarr_manifest_path", "zarr_manifest_md5sum", "split"}))

    def __repr__(self):
        return self.model_dump_json(exclude={"zarr_manifest_path", "zarr_manifest_md5sum", "split"}, indent=2)

    def __str__(self):
        return self.__repr__()



import pandas as pd

from polaris.dataset._subset import Subset
from polaris.evaluate import BenchmarkPredictions, BenchmarkResults, Metric
from polaris.utils.types import IncomingPredictionsType


def _optionally_subset(
    preds: BenchmarkPredictions | None,
    test_set_labels: list[str] | str,
    target_labels: list[str] | str,
) -> BenchmarkPredictions | None:
    """
    Returns the value in a nested dictionary associated with a sequence of keys
    if it exists, otherwise return None
    """
    if preds is None:
        return None

    if not isinstance(test_set_labels, list):
        test_set_labels = [test_set_labels]

    if not isinstance(target_labels, list):
        target_labels = [target_labels]

    return preds.get_subset(
        test_set_subset=test_set_labels,
        target_subset=target_labels,
    )


def evaluate_benchmark(
    target_cols: list[str],
    test_set_labels: list[str],
    test_set_sizes: dict[str, int],
    metrics: set[Metric],
    y_true: dict[str, Subset],
    y_pred: IncomingPredictionsType | None = None,
    y_prob: IncomingPredictionsType | None = None,
):
    """
    Utility function that contains the evaluation logic for a benchmark
    """

    # Normalize the and predictions to a consistent, internal representation.
    # Format is a two-level dictionary: {test_set_label: {target_label: np.ndarray}}
    if y_pred is not None:
        y_pred = BenchmarkPredictions(
            predictions=y_pred,
            target_labels=target_cols,
            test_set_labels=test_set_labels,
            test_set_sizes=test_set_sizes,
        )
    if y_prob is not None:
        y_prob = BenchmarkPredictions(
            predictions=y_prob,
            target_labels=target_cols,
            test_set_labels=test_set_labels,
            test_set_sizes=test_set_sizes,
        )

    # Compute the results
    # Results are saved in a tabular format. For more info, see the BenchmarkResults docs.
    scores = pd.DataFrame(columns=BenchmarkResults.RESULTS_COLUMNS)

    # For every test set...
    for test_label in test_set_labels:
        # For every metric...
        for metric in metrics:
            if metric.is_multitask:
                # Multi-task but with a metric across targets
                score = metric(
                    y_true=y_true[test_label],
                    y_pred=_optionally_subset(y_pred, test_set_labels=test_label),
                    y_prob=_optionally_subset(y_prob, test_set_labels=test_label),
                )

                scores.loc[len(scores)] = (test_label, "aggregated", metric, score)
                continue

            # Otherwise, for every target...
            for target_label in target_cols:
                score = metric(
                    y_true=y_true[test_label].filter_targets(target_label),
                    y_pred=_optionally_subset(y_pred, test_set_labels=test_label, target_labels=target_label),
                    y_prob=_optionally_subset(y_prob, test_set_labels=test_label, target_labels=target_label),
                )

                scores.loc[len(scores)] = (test_label, target_label, metric.name, score)

    return scores



"""
The code in this file is based on the zarr-checksum package

Mainted by Jacob Nesbitt, released under the DANDI org on Github
and with Kitware, Inc. credited as the author. This code is released
with the Apache 2.0 license.

See also: https://github.com/dandi/zarr_checksum

Instead of adding the package as a dependency, we opted to copy over the code
because it is a small and self-contained module that we will want to alter to
support our Polaris code base.

NOTE: We have made some modifications to the original code.

----

Copyright 2023 Kitware, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import uuid
from pathlib import Path
from shutil import copytree, rmtree

import pytest
import zarr

from polaris.dataset.zarr._checksum import (
    EMPTY_CHECKSUM,
    InvalidZarrChecksum,
    _ZarrChecksum,
    _ZarrChecksumManifest,
    _ZarrChecksumTree,
    _ZarrDirectoryDigest,
    compute_zarr_checksum,
)


def test_generate_digest() -> None:
    manifest = _ZarrChecksumManifest(
        directories=[_ZarrChecksum(digest="a7e86136543b019d72468ceebf71fb8e-1-1", name="a/b", size=1)],
        files=[_ZarrChecksum(digest="92eb5ffee6ae2fec3ad71c777531578f-0-1", name="b", size=1)],
    )
    assert manifest.generate_digest().digest == "9c5294e46908cf397cb7ef53ffc12efc-1-2"


def test_zarr_checksum_sort_order() -> None:
    # The a < b in the name should take precedence over z > y in the md5
    a = _ZarrChecksum(name="a", digest="z", size=3)
    b = _ZarrChecksum(name="b", digest="y", size=4)
    assert sorted([b, a]) == [a, b]


def test_parse_zarr_directory_digest() -> None:
    # Parse valid
    _ZarrDirectoryDigest.parse("c228464f432c4376f0de6ddaea32650c-37481-38757151179")
    _ZarrDirectoryDigest.parse(None)

    # Ensure exception is raised
    with pytest.raises(InvalidZarrChecksum):
        _ZarrDirectoryDigest.parse("asd")
    with pytest.raises(InvalidZarrChecksum):
        _ZarrDirectoryDigest.parse("asd-0--0")


def test_pop_deepest() -> None:
    tree = _ZarrChecksumTree()
    tree.add_leaf(Path("a/b"), size=1, digest="asd")
    tree.add_leaf(Path("a/b/c"), size=1, digest="asd")
    node = tree.pop_deepest()

    # Assert popped node is a/b/c, not a/b
    assert node.path == Path("a/b")
    assert len(node.checksums.files) == 1
    assert len(node.checksums.directories) == 0
    assert node.checksums.files[0].name == "c"


def test_process_empty_tree() -> None:
    tree = _ZarrChecksumTree()
    assert tree.process().digest == EMPTY_CHECKSUM


def test_process_tree() -> None:
    tree = _ZarrChecksumTree()
    tree.add_leaf(Path("a/b"), size=1, digest="9dd4e461268c8034f5c8564e155c67a6")
    tree.add_leaf(Path("c"), size=1, digest="415290769594460e2e485922904f345d")
    checksum = tree.process()

    # This zarr checksum was computed against the same file structure using the previous
    # zarr checksum implementation
    # Assert the current implementation produces a matching checksum
    assert checksum.digest == "e53fcb7b5c36b2f4647fbf826a44bdc9-2-2"


def test_checksum_for_zarr_archive(zarr_archive, tmp_path):
    # NOTE: This test was not in the original code base of the zarr-checksum package.
    checksum, _ = compute_zarr_checksum(zarr_archive)

    path = str(tmp_path / "copy")
    copytree(zarr_archive, path)
    assert checksum == compute_zarr_checksum(path)[0]

    root = zarr.open(path)
    root["A"][0:10] = 0
    assert checksum != compute_zarr_checksum(path)[0]


def test_zarr_leaf_to_checksum(zarr_archive):
    # NOTE: This test was not in the original code base of the zarr-checksum package.
    _, leaf_to_checksum = compute_zarr_checksum(zarr_archive)
    root = zarr.open(zarr_archive)

    # Check the basic structure - Each key corresponds to a file in the zarr archive
    assert len(leaf_to_checksum) == len(root.store)
    assert all(k.path in root.store for k in leaf_to_checksum)


def test_zarr_checksum_fails_for_remote_storage(zarr_archive):
    # NOTE: This test was not in the original code base of the zarr-checksum package.
    with pytest.raises(RuntimeError):
        compute_zarr_checksum("s3://bucket/data.zarr")
    with pytest.raises(RuntimeError):
        compute_zarr_checksum("gs://bucket/data.zarr")


def test_zarr_checksum_with_path_normalization(zarr_archive):
    # NOTE: This test was not in the original code base of the zarr-checksum package.

    baseline = compute_zarr_checksum(zarr_archive)[0]
    rootdir = os.path.dirname(zarr_archive)

    # Test a relative path
    copytree(zarr_archive, os.path.join(rootdir, "relative", "data.zarr"))
    compute_zarr_checksum(f"{zarr_archive}/../relative/data.zarr")[0] == baseline

    # Test with variables
    rng_id = str(uuid.uuid4())
    os.environ["TMP_TEST_DIR"] = rng_id
    copytree(zarr_archive, os.path.join(rootdir, "vars", rng_id))
    compute_zarr_checksum(f"{rootdir}/vars/${{TMP_TEST_DIR}}")[0] == baseline  # Format ${...}
    compute_zarr_checksum(f"{rootdir}/vars/$TMP_TEST_DIR")[0] == baseline  # Format $...

    # And with the user abbreviation
    try:
        path = os.path.expanduser("~/data.zarr")
        copytree(zarr_archive, path)
        compute_zarr_checksum("~/data.zarr")[0] == baseline
    finally:
        rmtree(path)



import zarr
import zarr.storage

from polaris.utils.errors import InvalidZarrCodec

try:
    # Register imagecodecs if they are available.
    from imagecodecs.numcodecs import register_codecs

    register_codecs()
except ImportError:
    pass


def load_zarr_group_to_memory(group: zarr.Group) -> dict:
    """Loads an entire Zarr group into memory."""

    if isinstance(group, dict):
        # If a Zarr group is already loaded to memory (e.g. with dataset.load_to_memory()),
        # the adapter would receive a dictionary instead of a Zarr group.
        return group

    data = {}
    for key, item in group.items():
        if isinstance(item, zarr.Array):
            data[key] = item[:]
        elif isinstance(item, zarr.Group):
            data[key] = load_zarr_group_to_memory(item)
    return data


def check_zarr_codecs(group: zarr.Group):
    """Check if all codecs in the Zarr group are registered."""
    try:
        for key, item in group.items():
            if isinstance(item, zarr.Group):
                check_zarr_codecs(item)

    except ValueError as error:
        # Zarr raises a generic ValueError if a codec is not registered.
        # See also: https://github.com/zarr-developers/zarr-python/issues/2508
        prefix = "codec not available: "
        error_message = str(error)

        if not error_message.startswith(prefix):
            raise error

        # Remove prefix and apostrophes
        codec_id = error_message.removeprefix(prefix).strip("'")
        raise InvalidZarrCodec(codec_id)



# This script includes docking related evaluation metrics.


import numpy as np
from rdkit.Chem.rdMolAlign import CalcRMS

import datamol as dm


def _rmsd(mol_probe: dm.Mol, mol_ref: dm.Mol) -> float:
    """Calculate RMSD between predicted molecule and closest ground truth molecule.
       The RMSD is calculated with first conformer of predicted molecule and only consider heavy atoms for RMSD calculation.
       It is assumed that the predicted binding conformers are extracted from the docking output, where the receptor (protein) coordinates have been aligned with the original crystal structure.

    Args:
        mol_probe: Predicted molecule (docked ligand) with exactly one conformer.
        mol_ref: Ground truth molecule (crystal ligand) with at least one conformer. If multiple conformers are
            present, the lowest RMSD will be reported.

    Returns:
        Returns the RMS between two molecules, taking symmetry into account.
    """

    # copy the molecule for modification.
    mol_probe = dm.copy_mol(mol_probe)
    mol_ref = dm.copy_mol(mol_ref)

    # remove hydrogen from molecule
    mol_probe = dm.remove_hs(mol_probe)
    mol_ref = dm.remove_hs(mol_ref)

    # calculate RMSD
    return CalcRMS(
        prbMol=mol_probe, refMol=mol_ref, symmetrizeConjugatedTerminalGroups=True, prbId=-1, refId=-1
    )


def rmsd_coverage(y_pred: str | list[dm.Mol], y_true: str | list[dm.Mol], max_rsmd: float = 2):
    """
    Calculate the coverage of molecules with an RMSD less than a threshold (2 Å by default) compared to the reference molecule conformer.

    It is assumed that the predicted binding conformers are extracted from the docking output, where the receptor (protein) coordinates have been aligned with the original crystal structure.

    Attributes:
        y_pred: List of predicted binding conformers.
        y_true: List of ground truth binding confoermers.
        max_rsmd: The threshold for determining acceptable rsmd.
    """

    if len(y_pred) != len(y_true):
        assert ValueError(
            f"The list of probing molecules and the list of reference molecules are different sizes. {len(y_pred)} != {len(y_true)} "
        )

    rmsds = np.array(
        [_rmsd(mol_probe=mol_probe, mol_ref=mol_ref) for mol_probe, mol_ref in zip(y_pred, y_true)]
    )

    return np.sum(rmsds <= max_rsmd) / len(rmsds)



import json
from hashlib import md5
from os import PathLike
from pathlib import Path
from typing import Any, ClassVar, Literal

import fsspec
import numpy as np
import pandas as pd
import zarr
from datamol.utils import fs as dmfs
from pydantic import PrivateAttr, computed_field, field_validator, model_validator
from typing_extensions import Self, deprecated

from polaris.dataset._adapters import Adapter
from polaris.dataset._base import BaseDataset
from polaris.dataset.zarr import ZarrFileChecksum, compute_zarr_checksum
from polaris.mixins._checksum import ChecksumMixin
from polaris.utils.errors import InvalidDatasetError
from polaris.utils.types import (
    AccessType,
    ChecksumStrategy,
    HubOwner,
    ZarrConflictResolution,
)

# Constants
_SUPPORTED_TABLE_EXTENSIONS = ["parquet"]
_INDEX_SEP = "#"


@deprecated(
    "Use DatasetV2 instead. If you're loading this dataset from the Polaris Hub, you can ignore this warning."
)
class DatasetV1(BaseDataset, ChecksumMixin):
    """First version of a Polaris Dataset.

    Stores datapoints in a Pandas DataFrame and implements _pointer columns_ to support the storage of XL data
    outside the DataFrame in a Zarr archive.

    Info: Pointer columns
        For complex data, such as images, we support storing the content in external blobs of data.
        In that case, the table contains _pointers_ to these blobs that are dynamically loaded when needed.

    Attributes:
        table: The core data-structure, storing data-points in a row-wise manner. Can be specified as either a
            path to a `.parquet` file or a `pandas.DataFrame`.

    For additional metadata attributes, see the base classes.

    Raises:
        InvalidDatasetError: If the dataset does not conform to the Pydantic data-model specification.
    """

    _artifact_type = "dataset"
    _version: ClassVar[Literal[1]] = 1

    # Public attributes
    # Data
    table: pd.DataFrame

    # Private attributes
    _zarr_md5sum_manifest: list[ZarrFileChecksum] = PrivateAttr(default_factory=list)

    @field_validator("table", mode="before")
    @classmethod
    def _load_table(cls, v) -> pd.DataFrame:
        """
        Load from path if not a dataframe
        """
        if isinstance(v, str):
            if not dmfs.is_file(v) or dmfs.get_extension(v) not in _SUPPORTED_TABLE_EXTENSIONS:
                raise InvalidDatasetError(f"{v} is not a valid DataFrame or .parquet path.")
            v = pd.read_parquet(v)
        return v

    @field_validator("table")
    @classmethod
    def _validate_table(cls, v: pd.DataFrame) -> pd.DataFrame:
        """
        Make sure that the pandas index is contiguous and starts at 0, and
        that all columns are named and unique.
        """
        # Check if there are any duplicate columns
        if any(v.columns.duplicated()):
            raise InvalidDatasetError("The table contains duplicate columns")
        # Check if there are any unnamed columns
        if not all(isinstance(c, str) for c in v.columns):
            raise InvalidDatasetError("The table contains unnamed columns")
        return v

    @model_validator(mode="after")
    def _validate_v1_dataset_model(self) -> Self:
        """Verifies some dependencies between properties"""

        has_pointers = any(anno.is_pointer for anno in self.annotations.values())
        if has_pointers and self.zarr_root_path is None:
            raise InvalidDatasetError("A zarr_root_path needs to be specified when there are pointer columns")
        if not has_pointers and self.zarr_root_path is not None:
            raise InvalidDatasetError(
                "The zarr_root_path should only be specified when there are pointer columns"
            )

        return self

    @field_validator("default_adapters", mode="before")
    def _validate_adapters(cls, value):
        """Validate the adapters"""
        return {k: Adapter[v] if isinstance(v, str) else v for k, v in value.items()}

    def _compute_checksum(self) -> str:
        """Computes a hash of the dataset.

        This is meant to uniquely identify the dataset and can be used to verify the version.

        1. Is not sensitive to the ordering of the columns or rows in the table.
        2. Purposefully does not include the metadata (source, description, name, annotations).
        3. Includes a hash for the Zarr archive.
        """
        hash_fn = md5()

        # Sort the columns s.t. the checksum is not sensitive to the column-ordering
        df = self.table.copy(deep=True)
        df = df[sorted(df.columns.tolist())]

        # Use the sum of the row-wise hashes s.t. the hash is insensitive to the row-ordering
        table_hash = pd.util.hash_pandas_object(df, index=False).sum()
        hash_fn.update(table_hash)

        # If the Zarr archive exists, we hash its contents too.
        if self.uses_zarr:
            zarr_hash, self._zarr_md5sum_manifest = compute_zarr_checksum(self.zarr_root_path)
            hash_fn.update(zarr_hash.digest.encode())

        checksum = hash_fn.hexdigest()
        return checksum

    def load_zarr_root_from_hub(self):
        """
        Loads a Zarr archive from the Hub.
        """
        from polaris.hub.client import PolarisHubClient
        from polaris.hub.storage import StorageSession

        with PolarisHubClient() as client:
            with StorageSession(client, "read", self.urn) as storage:
                return zarr.open_consolidated(store=storage.store("extension"))

    @computed_field
    @property
    def zarr_md5sum_manifest(self) -> list[ZarrFileChecksum]:
        """
        The Zarr Checksum manifest stores the checksums of all files in a Zarr archive.
        If the dataset doesn't use Zarr, this will simply return an empty list.
        """
        if len(self._zarr_md5sum_manifest) == 0 and not self.has_md5sum:
            # The manifest is set as an instance variable
            # as a side effect of the compute_checksum method
            self.md5sum = self._compute_checksum()
        return self._zarr_md5sum_manifest

    @property
    def rows(self) -> list[str | int]:
        """Return all row indices for the dataset"""
        return self.table.index.tolist()

    @property
    def columns(self) -> list[str]:
        """Return all columns for the dataset"""
        return self.table.columns.tolist()

    @property
    def dtypes(self) -> dict[str, np.dtype]:
        """Return the dtype for each of the columns for the dataset"""
        return {col: self.table[col].dtype for col in self.columns}

    def get_data(
        self, row: str | int, col: str, adapters: dict[str, Adapter] | None = None
    ) -> np.ndarray | Any:
        """Since the dataset might contain pointers to external files, data retrieval is more complicated
        than just indexing the `table` attribute. This method provides an end-point for seamlessly
        accessing the underlying data.

        Args:
            row: The row index in the `Dataset.table` attribute
            col: The column index in the `Dataset.table` attribute
            adapters: The adapters to apply to the data before returning it.
                If None, will use the default adapters specified for the dataset.

        Returns:
            A numpy array with the data at the specified indices. If the column is a pointer column,
                the content of the referenced file is loaded to memory.
        """

        # Fetch adapters for dataset and a given column
        # Partially override if the adapters parameter is specified.
        adapters = {**self.default_adapters, **(adapters or {})}
        adapter = adapters.get(col)

        # If not a pointer, return it here. Apply adapter if specified.
        value = self.table.at[row, col]
        if not self.annotations[col].is_pointer:
            if adapter is not None:
                return adapter(value)
            return value

        # Load the data from the Zarr archive
        path, index = self._split_index_from_path(value)
        arr = self.zarr_data[path][index]

        # Change to tuple if a slice
        if isinstance(index, slice):
            arr = tuple(arr)

        # Adapt the input to the specified format
        if adapter is not None:
            arr = adapter(arr)

        return arr

    def upload_to_hub(
        self,
        access: AccessType = "private",
        owner: HubOwner | str | None = None,
        parent_artifact_id: str | None = None,
    ):
        """
        Very light, convenient wrapper around the
        [`PolarisHubClient.upload_dataset`][polaris.hub.client.PolarisHubClient.upload_dataset] method.
        """
        from polaris.hub.client import PolarisHubClient

        with PolarisHubClient() as client:
            client.upload_dataset(self, owner=owner, access=access, parent_artifact_id=parent_artifact_id)

    @classmethod
    def from_json(cls, path: str):
        """Loads a dataset from a JSON file.

        Args:
            path: The path to the JSON file to load the dataset from .ColumnAnnotation
        """
        with fsspec.open(path, "r") as f:
            data = json.load(f)
        return cls.model_validate(data)

    def _split_index_from_path(self, path: str) -> tuple[str, int | None]:
        """
        Paths can have an additional index appended to them.
        This extracts that index from the path.
        """
        index = None
        if _INDEX_SEP in path:
            path, index = path.split(_INDEX_SEP)
            index = index.split(":")

            if len(index) == 1:
                index = int(index[0])
            elif len(index) == 2:
                index = slice(int(index[0]), int(index[1]))
            else:
                raise ValueError(f"Invalid index format: {index}")
        return path, index

    def _repr_dict_(self) -> dict:
        """Utility function for pretty-printing to the command line and jupyter notebooks"""
        repr_dict = self.model_dump(exclude={"table", "zarr_md5sum_manifest", "md5sum"})
        return repr_dict

    def __eq__(self, other):
        """Whether two datasets are equal is solely determined by the checksum"""
        if not isinstance(other, DatasetV1):
            return False
        return self.md5sum == other.md5sum

    def _table_serialization_path(self, destination: Path) -> Path:
        return destination / "table.parquet"

    def to_json(self, destination: str | Path, if_exists: ZarrConflictResolution = "replace") -> str:
        """
        Save the dataset to a destination directory as a JSON file.

        Warning: Multiple files
            Perhaps unintuitive, this method creates multiple files.

            1. `/path/to/destination/[dataset.slug].json`: This file can be loaded with `Dataset.from_json`.
            2. `/path/to/destination/table.parquet`: The `Dataset.table` attribute is saved here.
            3. _(Optional)_ `/path/to/destination/[dataset.zarr_root]`: Any additional blobs of data referenced by the
                    pointer columns will be stored here.

        Args:
            destination: The _directory_ to save the associated data to.
            if_exists: Action for handling existing files in the Zarr archive. Options are 'raise' to throw
                an error, 'replace' to overwrite, or 'skip' to proceed without altering the existing files.

        Returns:
            The path to the JSON file.
        """
        destination = Path(destination)
        destination.mkdir(exist_ok=True, parents=True)

        # Make a copy to cache the data, and then serialize that
        copy = self.model_copy()
        copy.cache(destination, if_exists=if_exists)

        serialized = copy.model_dump(exclude_none=True, exclude={"zarr_manifest_path", "table"})
        serialized["table"] = str(copy._table_serialization_path(destination))

        destination_json = destination / f"{copy.slug}.json"
        destination_json.write_text(json.dumps(serialized))
        return str(destination_json)

    def cache(
        self,
        destination: str | PathLike | None = None,
        if_exists: ZarrConflictResolution = "replace",
        verify_checksum: bool = True,
    ) -> str:
        """Caches the dataset by downloading all additional data for pointer columns to a local directory.

        Args:
            destination: The directory to cache the data to. If None, will use the default cache directory.
            if_exists: Action for handling existing files at the destination. Options are 'raise' to throw
                an error, 'replace' to overwrite, or 'skip' to proceed without altering the existing files.
            verify_checksum: Whether to verify the checksum of the dataset after caching.

        Returns:
            The path to the directory where data has been cached to.
        """
        if not destination:
            destination = self._cache_dir

        destination = Path(destination)
        destination.mkdir(exist_ok=True, parents=True)

        table_path = self._table_serialization_path(destination)
        self.table.to_parquet(table_path)

        self._cache_zarr(destination, if_exists)

        if verify_checksum:
            self.verify_checksum()
        else:
            self._md5sum = None

        return str(destination)

    def should_verify_checksum(self, strategy: ChecksumStrategy) -> bool:
        """
        Determines whether to verify the checksum of the dataset based on the strategy.
        """
        match strategy:
            case "ignore":
                return False
            case "verify":
                return True
            case "verify_unless_zarr":
                return not self.uses_zarr



from typing import ClassVar

import pandas as pd
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    computed_field,
    field_serializer,
    field_validator,
    model_validator,
)
from pydantic.alias_generators import to_camel

from polaris.evaluate import ResultsMetadataV1, ResultsMetadataV2
from polaris.utils.errors import InvalidResultError
from polaris.utils.misc import slugify
from polaris.utils.types import (
    AccessType,
    HubOwner,
    SlugCompatibleStringType,
)

# Define some helpful type aliases
TestLabelType = str
TargetLabelType = str


class ResultRecords(BaseModel):
    """
    A grouped, tabular data-structure to save the actual data of the benchmark results.
    The grouping by test set and target label is done to make it easier to build leaderboards.
    """

    test_set: TestLabelType
    target_label: TargetLabelType
    scores: dict[str, float]

    # Model config
    model_config = ConfigDict(alias_generator=to_camel, populate_by_name=True)


class BaseEvaluationResult:
    """Base class for saving evaluation results

    The actual results are saved in the `results` field using the following tabular format:

    | Test set | Target label | Metric | Score |
    | -------- | ------------ | ------ | ----- |
    | test_iid | EGFR_WT      | AUC    | 0.9   |
    | test_ood | EGFR_WT      | AUC    | 0.75  |
    |    ...   |      ...     |  ...   |  ...  |
    | test_ood | EGFR_L858R   | AUC    | 0.79  |

    question: Categorizing methods
        An open question is how to best categorize a methodology (e.g. a model).
        This is needed since we would like to be able to aggregate results across benchmarks too,
        to say something about which (type of) methods performs best _in general_.

    Attributes:
        results: Evaluation results are stored directly in a dataframe or in a serialized, JSON compatible dict
            that can be decoded into the associated tabular format.
    For additional metadata attributes, see the base classes.
    """

    # Define the columns of the results table
    RESULTS_COLUMNS: ClassVar[list[str]] = ["Test set", "Target label", "Metric", "Score"]

    # Results attribute
    results: pd.DataFrame

    @field_validator("results", mode="before")
    @classmethod
    def _convert_results(cls, v: pd.DataFrame | list[ResultRecords | dict]) -> pd.DataFrame:
        """
        Convert the results to a dataframe if they are not already in that format
        """
        if isinstance(v, pd.DataFrame):
            return v

        try:
            df = pd.DataFrame(columns=cls.RESULTS_COLUMNS)
            for record in v:
                if isinstance(record, dict):
                    record = ResultRecords(**record)

                for metric, score in record.scores.items():
                    df.loc[len(df)] = {
                        "Test set": record.test_set,
                        "Target label": record.target_label,
                        "Metric": metric,
                        "Score": score,
                    }
            return df

        except (ValueError, UnicodeDecodeError) as error:
            raise InvalidResultError(
                f"The provided dictionary cannot be parsed into a {cls.__name__} instance."
            ) from error

    @field_validator("results")
    @classmethod
    def _validate_results(cls, v: pd.DataFrame) -> pd.DataFrame:
        """
        Ensure the results are a valid dataframe and have the expected columns
        """
        # Check if the dataframe contains _only_ the expected columns
        if set(v.columns) != set(cls.RESULTS_COLUMNS):
            raise InvalidResultError(
                f"The results dataframe should have the following columns: {cls.RESULTS_COLUMNS}"
            )

        # Check if the results are not empty
        if v.empty:
            raise InvalidResultError("The results dataframe is empty")

        # NOTE (cwognum): Since we have a reference to the benchmark, I considered validating the values in the
        #  columns as well (e.g. are all metrics, targets and test sets actually part of the benchmark).
        #  However, to keep this class light-weight, I did not want to add a strict dependency on the full benchmark class.
        #  Especially because validation will happen on the Hub as well before it is shown there.
        return v

    @field_serializer("results")
    def _serialize_results(self, value: pd.DataFrame) -> list[ResultRecords]:
        """Change from the Metric enum to a string representation"""
        serialized = []
        grouped = value.groupby(["Test set", "Target label"])
        for (test_set, target_label), group in grouped:
            metrics = {row["Metric"]: row["Score"] for _, row in group.iterrows()}
            record = ResultRecords(test_set=test_set, target_label=target_label, scores=metrics)
            serialized.append(record)

        return serialized


class EvaluationResultV1(ResultsMetadataV1, BaseEvaluationResult):
    """V1 implementation of evaluation results without model field support"""

    pass


class EvaluationResultV2(ResultsMetadataV2, BaseEvaluationResult):
    """V2 implementation of evaluation results with model field replacing URLs"""

    pass


class BaseBenchmarkResults:
    """Base class for results of standard benchmarks.

    This object is returned by  `benchmark.evaluate()`.

    In addition to the metrics on the test set, it contains additional metadata and logic to integrate
    the results with the Polaris Hub.

    benchmark_name: The name of the benchmark for which these results were generated.
        Together with the benchmark owner, this uniquely identifies the benchmark on the Hub.
    benchmark_owner: The owner of the benchmark for which these results were generated.
        Together with the benchmark name, this uniquely identifies the benchmark on the Hub.
    """

    _artifact_type = "result"

    benchmark_artifact_id: str | None = Field(None)
    benchmark_name: SlugCompatibleStringType | None = Field(None, deprecated=True)
    benchmark_owner: HubOwner | None = Field(None, deprecated=True)

    @model_validator(mode="after")
    def set_benchmark_artifact_id(self):
        if self.benchmark_artifact_id is None and self.benchmark_name and self.benchmark_owner:
            self.benchmark_artifact_id = f"{self.benchmark_owner}/{slugify(self.benchmark_name)}"
        return self

    def upload_to_hub(
        self,
        access: AccessType = "private",
        owner: HubOwner | str | None = None,
        **kwargs: dict,
    ):
        """
        Very light, convenient wrapper around the
        [`PolarisHubClient.upload_results`][polaris.hub.client.PolarisHubClient.upload_results] method.
        """
        from polaris.hub.client import PolarisHubClient

        with PolarisHubClient(**kwargs) as client:
            return client.upload_results(self, access=access, owner=owner)


class BenchmarkResultsV1(EvaluationResultV1, BaseBenchmarkResults):
    """V1 implementation of benchmark results without model field support"""

    pass


class BenchmarkResultsV2(EvaluationResultV2, BaseBenchmarkResults):
    """V2 implementation of benchmark results with model field replacing URLs"""

    pass


class CompetitionResults(EvaluationResultV1):
    """Class specific to results for competition benchmarks.

    This object is returned by [`CompetitionSpecification.evaluate`][polaris.competition.CompetitionSpecification.evaluate].
    In addition to the metrics on the test set, it contains additional metadata and logic to integrate
    the results with the Polaris Hub.

    Attributes:
        competition_name: The name of the competition for which these results were generated.
            Together with the competition owner, this uniquely identifies the competition on the Hub.
        competition_owner: The owner of the competition for which these results were generated.
            Together with the competition name, this uniquely identifies the competition on the Hub.
    """

    _artifact_type = "competitionResult"

    competition_name: SlugCompatibleStringType = Field(..., frozen=True)
    competition_owner: HubOwner | None = Field(None, frozen=True)

    @computed_field
    @property
    def competition_artifact_id(self) -> str:
        return f"{self.competition_owner}/{slugify(self.competition_name)}"



import logging
from functools import cached_property
from hashlib import md5
from typing import Generator, Sequence

from pydantic import BaseModel, ConfigDict, Field, computed_field, field_validator, model_validator
from pydantic.alias_generators import to_camel
from pyroaring import BitMap
from typing_extensions import Self

from polaris.utils.errors import InvalidBenchmarkError

logger = logging.getLogger(__name__)


class IndexSet(BaseModel):
    """
    A set of indices for a split, either training or test.

    It wraps a Roaring Bitmap object to store the indices, and provides
    useful properties when serializing for upload to the Hub.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, alias_generator=to_camel)

    indices: BitMap = Field(default_factory=BitMap, frozen=True, exclude=True)

    @field_validator("indices", mode="before")
    @classmethod
    def _validate_indices(cls, v: BitMap | Sequence[int]) -> BitMap:
        """
        Accepts an initial sequence of ints, and turn it into a BitMap
        """
        if isinstance(v, BitMap):
            return v
        return BitMap(v)

    @computed_field
    @cached_property
    def datapoints(self) -> int:
        return len(self.indices)

    @computed_field
    @cached_property
    def md5_checksum(self) -> str:
        return md5(self.serialize()).hexdigest()

    def intersect(self, other: Self) -> bool:
        return self.indices.intersect(other.indices)

    def serialize(self) -> bytes:
        return self.indices.serialize()

    @staticmethod
    def deserialize(index_set: bytes) -> "IndexSet":
        return IndexSet(indices=BitMap.deserialize(index_set))


class SplitV2(BaseModel):
    training: IndexSet
    test: IndexSet

    @field_validator("training", "test", mode="before")
    @classmethod
    def _parse_index_sets(cls, v: bytes | IndexSet) -> bytes | IndexSet:
        """
        Accepted a binary serialized IndexSet
        """
        if isinstance(v, bytes):
            return IndexSet.deserialize(v)
        return v

    @field_validator("training")
    @classmethod
    def _validate_training_set(cls, v: IndexSet) -> IndexSet:
        """
        Training index set can be empty (zero-shot)
        """
        if v.datapoints == 0:
            logger.info(
                "This benchmark only specifies a test set. It will return an empty train set in `get_train_test_split()`"
            )
        return v

    @field_validator("test")
    @classmethod
    def _validate_test_set(cls, v: IndexSet) -> IndexSet:
        """
        Test index set cannot be empty
        """
        if v.datapoints == 0:
            raise InvalidBenchmarkError("The predefined split contains empty test partitions")
        return v

    @model_validator(mode="after")
    def validate_set_overlap(self) -> Self:
        """
        The training and test index sets do not overlap
        """
        if self.training.intersect(self.test):
            raise InvalidBenchmarkError("The predefined split specifies overlapping train and test sets")
        return self

    @property
    def n_train_datapoints(self) -> int:
        """
        The size of the train set.
        """
        return self.training.datapoints

    @property
    def n_test_sets(self) -> int:
        """
        The number of test sets
        """
        # TODO: Until we support multi-test benchmarks
        return 1

    @property
    def n_test_datapoints(self) -> dict[str, int]:
        """
        The size of (each of) the test set(s).
        """
        # TODO: Until we support multi-test benchmarks
        return {"test": self.test.datapoints}

    @property
    def max_index(self) -> int:
        # TODO: Until we support multi-test benchmarks (need)
        return max(self.training.indices.max(), self.test.indices.max())

    def test_items(self) -> Generator[tuple[str, IndexSet], None, None]:
        # TODO: Until we support multi-test benchmarks
        yield "test", self.test


class SplitSpecificationV2Mixin(BaseModel):
    """
    Mixin class to add a split field to a benchmark. This is the V2 implementation.

    The internal representation for the split is a roaring bitmap,
    which drastically improves scalability over the V1 implementation.

    Attributes:
        split: The predefined train-test split to use for evaluation.
    """

    split: SplitV2

    @computed_field
    @property
    def n_train_datapoints(self) -> int:
        """The size of the train set."""
        return self.split.n_train_datapoints

    @computed_field
    @property
    def n_test_sets(self) -> int:
        """The number of test sets"""
        return self.split.n_test_sets

    @computed_field
    @property
    def n_test_datapoints(self) -> dict[str, int]:
        """The size of (each of) the test set(s)."""
        return self.split.n_test_datapoints

    @computed_field
    @property
    def test_set_sizes(self) -> dict[str, int]:
        return {label: index_set.datapoints for label, index_set in self.split.test_items()}

    @computed_field
    @property
    def test_set_labels(self) -> list[str]:
        return list(label for label, _ in self.split.test_items())



from pathlib import Path
from typing import TYPE_CHECKING, Optional, Sequence

import fastpdb
import numpy as np
import pandas as pd
import zarr
from fastpdb import struc
from typing_extensions import deprecated

from polaris.dataset import ColumnAnnotation, Modality
from polaris.dataset._adapters import Adapter
from polaris.dataset.converters._base import Converter, FactoryProduct
from polaris.dataset.zarr._utils import load_zarr_group_to_memory

if TYPE_CHECKING:
    from polaris.dataset import DatasetFactory


DTYPE_DICT = {
    "chain_id": "U4",
    "res_id": int,
    "ins_code": "U1",
    "res_name": "U5",
    "hetero": bool,
    "atom_name": "U6",
    "element": "U2",
}

KEYS = [
    "X",
    "Y",
    "Z",
    "chain_id",
    "res_id",
    "ins_code",
    "res_name",
    "hetero",
    "atom_name",
    "element",
    "atom_id",
    "b_factor",
    "occupancy",
    "charge",
]


def zarr_to_pdb(atom_dict: zarr.Group):
    """Load a dictionary of arrays to fastpdb AtomArray"""

    atom_array = []

    # Load to memory once to drastically speed up the conversion
    # Otherwise you keep decompressing and copying entire chunks of data to just use a single row
    atom_dict = load_zarr_group_to_memory(atom_dict)

    # convert dictionary to array list of Atom object
    array_length = atom_dict["X"].shape[0]
    for ind in range(array_length):
        atom = struc.Atom(
            coord=(atom_dict["X"][ind], atom_dict["Y"][ind], atom_dict["Z"][ind]),
            chain_id=atom_dict["chain_id"][ind],
            res_id=atom_dict["res_id"][ind],
            ins_code=atom_dict["ins_code"][ind],
            res_name=atom_dict["res_name"][ind],
            hetero=atom_dict["hetero"][ind],
            atom_name=atom_dict["atom_name"][ind],
            element=atom_dict["element"][ind],
            b_factor=atom_dict["b_factor"][ind],
            occupancy=atom_dict["occupancy"][ind],
            charge=atom_dict["charge"][ind],
            atom_id=atom_dict["atom_id"][ind],
        )
        atom_array.append(atom)

    # Note that this is a `fastpdb` AtomArray, not a NumPy array.
    return struc.array(atom_array)


@deprecated("Please use the custom codecs in `polaris.dataset.zarr.codecs` instead.")
class PDBConverter(Converter):
    """
    Converts PDB files into a Polaris dataset based on fastpdb.

    Info: Only the most essential structural information of a protein is retained
        This conversion saves the 3D coordinates, chain ID, residue ID, insertion code, residue name, heteroatom indicator, atom name, element, atom ID, B-factor, occupancy, and charge.
        Records such as CONECT (connectivity information), ANISOU (anisotropic Temperature Factors), HETATM (heteroatoms and ligands) are handled by `fastpdb`.
        We believe this makes for a good _ML-ready_ format, but let us know if you require any other information to be saved.


    Info: PDBs as ND-arrays using `biotite`
        To save PDBs in a Polaris-compatible format, we convert them to ND-arrays using `fastpdb` and `biotite`.
        We then save these ND-arrays to Zarr archives.
        For more info, see [fastpdb](https://github.com/biotite-dev/fastpdb)
        and [biotite](https://github.com/biotite-dev/biotite/blob/main/src/biotite/structure/atoms.py)

    Args:
        pdb_column: The name of the column that will contain the pointers to the pdbs.
        n_jobs: The number of jobs to run in parallel.
        zarr_chunks: The chunk size for the Zarr arrays.
    """

    def __init__(
        self,
        pdb_column: str = "pdb",
        n_jobs: int = 1,
        zarr_chunks: Sequence[Optional[int]] = (1,),
    ) -> None:
        super().__init__()
        self.pdb_column = pdb_column
        self.n_jobs = n_jobs
        self.chunks = zarr_chunks

    def _pdb_to_dict(self, atom_array: np.ndarray):
        """
        Convert 'biotite.AtomArray' to dictionary
        """
        pdb_dict = {k: [] for k in KEYS}
        for row in atom_array:
            pdb_dict["X"].append(row.coord[0])
            pdb_dict["Y"].append(row.coord[1])
            pdb_dict["Z"].append(row.coord[2])
            pdb_dict["chain_id"].append(row.chain_id)
            pdb_dict["res_id"].append(row.res_id)
            pdb_dict["ins_code"].append(row.ins_code)
            pdb_dict["res_name"].append(row.res_name)
            pdb_dict["hetero"].append(row.hetero)
            pdb_dict["atom_name"].append(row.atom_name)
            pdb_dict["element"].append(row.element)
            pdb_dict["atom_id"].append(row.atom_id)
            pdb_dict["b_factor"].append(row.b_factor)
            pdb_dict["occupancy"].append(row.occupancy)
            pdb_dict["charge"].append(row.charge)
        return pdb_dict

    def _load_pdb(self, path: str, pdb_pointer=None) -> dict:
        """
        Load a single PDB file along with a dictionary that contains structure properties as keys.

        `AtomArray` stores data for an entire structure model containing *n* atoms.
        See [AtomArray](https://github.com/biotite-dev/biotite/blob/0404084283765baef765cc869d86ee07a898d82f/src/biotite/structure/atoms.py#L558) for more details.
        """
        # load pdb file
        in_file = fastpdb.PDBFile.read(path)

        # get the structure as AtomArray
        atom_array = in_file.get_structure(
            model=1,
            include_bonds=True,
            extra_fields=["atom_id", "b_factor", "occupancy", "charge"],
        )

        # convert AtomArray to dict
        pdb_dict = self._pdb_to_dict(atom_array)

        return pdb_dict

    def _convert_pdb(
        self, path: str, factory: "DatasetFactory", pdb_pointer: str | int, append: bool = False
    ) -> FactoryProduct:
        """
        Convert a single pdb to zarr file
        """

        # load pdb as fastpbd and convert to dict
        pdb_dict = self._load_pdb(path)

        # Create group and add
        if append:
            if self.pdb_column not in factory.zarr_root:
                raise RuntimeError(
                    f"Group {self.pdb_column} doesn't exist in {factory.zarr_root}. \
                    Please make sure the Group {self.pdb_column} is created. Or set `append` to `False`."
                )
            else:
                pdb_group = factory.zarr_root[self.pdb_column]
        else:
            pdb_group = factory.zarr_root.create_group(self.pdb_column)

        group = pdb_group.create_group(pdb_pointer)
        for col_name, col_val in pdb_dict.items():
            col_val = np.array(col_val)
            # get the accepted dtype
            dtype = DTYPE_DICT.get(col_name, col_val.dtype)
            group.array(col_name, data=col_val, dtype=dtype)

    def convert(self, path, factory: "DatasetFactory", append: bool = False) -> FactoryProduct:
        """Convert one or a list of PDB files into Zarr"""
        pdb_pointers = []

        # load single pdb and convert to zarr
        pdb_pointer = Path(path).stem
        self._convert_pdb(path, factory, pdb_pointer, append)
        pdb_pointers.append(pdb_pointer)

        # Add a pointer column (path) to the table
        pointers = [f"{self.pdb_column}/{pointer}" for pointer in pdb_pointers]
        df = pd.DataFrame()
        df[self.pdb_column] = pd.Series(pointers)

        # Set the annotations
        annotations = {
            self.pdb_column: ColumnAnnotation(
                is_pointer=True, modality=Modality.PROTEIN_3D, content_type="chemical/x-pdb"
            )
        }

        # Return the dataframe and the annotations
        return df, annotations, {self.pdb_column: Adapter.ARRAY_TO_PDB}



from polaris._artifact import BaseArtifactModel
from polaris.utils.types import HttpUrlString
from polaris.utils.types import AccessType, HubOwner
from pydantic import Field


class Model(BaseArtifactModel):
    """
    Represents a Model artifact in the Polaris ecosystem.

    A Model artifact serves as a centralized representation of a method or model, encapsulating its metadata.
    It can be associated with multiple result artifacts but is immutable after upload, except for the README field.

    Examples:
        Basic API usage:
        ```python
        from polaris.model import Model

        # Create a new Model Card
        model = Model(
            name="MolGPS",
            description="Graph transformer foundation model for molecular modeling",
            code_url="https://github.com/datamol-io/graphium"
        )

        # Upload the model card to the Hub
        model.upload_to_hub(owner="recursion")
        ```

    Attributes:
        readme (str): A detailed README describing the model.
        code_url (HttpUrlString | None): Optional URL pointing to the model's code repository.
        report_url (HttpUrlString | None): Optional URL linking to a report or publication related to the model.
        artifact_version: The version of the model.
        artifact_changelog: A description of the changes made in this model version.

    Methods:
        upload_to_hub(access: AccessType = "private", owner: HubOwner | str | None = None):
            Uploads the model artifact to the Polaris Hub, associating it with a specified owner and access level.

    For additional metadata attributes, see the base class.
    """

    _artifact_type = "model"

    readme: str = ""
    code_url: HttpUrlString | None = None
    report_url: HttpUrlString | None = None

    # Version-related fields
    artifact_version: int = Field(default=1, frozen=True)
    artifact_changelog: str | None = None

    def upload_to_hub(
        self,
        access: AccessType = "private",
        owner: HubOwner | str | None = None,
        parent_artifact_id: str | None = None,
    ):
        """
        Uploads the model to the Polaris Hub.
        """
        from polaris.hub.client import PolarisHubClient

        with PolarisHubClient() as client:
            client.upload_model(self, owner=owner, access=access, parent_artifact_id=parent_artifact_id)



import abc
import json
from hashlib import md5
from itertools import chain
from pathlib import Path
from typing import Any, Callable, ClassVar, Literal, TypeVar, Generic

import fsspec
import numpy as np
from pydantic import (
    BaseModel,
    Field,
    computed_field,
    field_validator,
    model_validator,
)
from sklearn.utils.multiclass import type_of_target
from typing_extensions import Self, deprecated

from polaris._artifact import BaseArtifactModel
from polaris.benchmark._split import SplitSpecificationV1Mixin
from polaris.benchmark._task import PredictiveTaskSpecificationMixin
from polaris.dataset import DatasetV1, Subset
from polaris.dataset._base import BaseDataset
from polaris.evaluate import BenchmarkResultsV1, BenchmarkResultsV2
from polaris.evaluate.utils import evaluate_benchmark
from polaris.hub.settings import PolarisHubSettings
from polaris.mixins import ChecksumMixin
from polaris.utils.dict2html import dict2html
from polaris.utils.errors import InvalidBenchmarkError
from polaris.utils.types import (
    AccessType,
    HubOwner,
    IncomingPredictionsType,
    TargetType,
)

# Type variable for the return type of evaluate
BenchmarkResultsType = TypeVar("BenchmarkResultsType", BenchmarkResultsV1, BenchmarkResultsV2)


class BaseSplitSpecificationMixin(BaseModel):
    """Base mixin class to add a split field to a benchmark."""

    split: Any

    @property
    @abc.abstractmethod
    def test_set_sizes(self) -> dict[str, int]:
        """The sizes of the test sets."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def n_test_sets(self) -> int:
        """The number of test sets"""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def n_train_datapoints(self) -> int:
        """The size of the train set."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def test_set_labels(self) -> list[str]:
        """The labels of the test sets."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def n_test_datapoints(self) -> dict[str, int]:
        """The size of (each of) the test set(s)."""
        raise NotImplementedError


class BenchmarkSpecification(
    PredictiveTaskSpecificationMixin,
    BaseArtifactModel,
    BaseSplitSpecificationMixin,
    abc.ABC,
    Generic[BenchmarkResultsType],
):
    """This class wraps a dataset with additional data to specify the evaluation logic.

    Specifically, it specifies:

    1. Which dataset to use;
    2. A task definition (we currently only support predictive tasks);
    3. A predefined, static train-test split to use during evaluation.

    Examples:
        Basic API usage:
        ```python
        import polaris as po

        # Load the benchmark from the Hub
        benchmark = po.load_benchmark("polaris/hello-world-benchmark")

        # Get the train and test data-loaders
        train, test = benchmark.get_train_test_split()

        # Use the training data to train your model
        # Get the input as an array with 'train.inputs' and 'train.targets'
        # Or simply iterate over the train object.
        for x, y in train:
            ...

        # Work your magic to accurately predict the test set
        predictions = [0.0 for x in test]

        # Evaluate your predictions
        results = benchmark.evaluate(predictions)

        # Submit your results
        results.upload_to_hub(owner="dummy-user")
        ```

    Attributes:
        dataset: The dataset the benchmark specification is based on.
        readme: Markdown text that can be used to provide a formatted description of the benchmark.
            If using the Polaris Hub, it is worth noting that this field is more easily edited through the Hub UI
            as it provides a rich text editor for writing markdown.
        artifact_version: The version of the benchmark.
        artifact_changelog: A description of the changes made in this benchmark version.
    For additional metadata attributes, see the base classes.
    """

    _artifact_type = "benchmark"

    dataset: BaseDataset = Field(exclude=True)
    readme: str = ""

    # Version-related fields
    artifact_version: int = Field(default=1, frozen=True)
    artifact_changelog: str | None = None

    @computed_field
    @property
    def dataset_artifact_id(self) -> str:
        return self.dataset.artifact_id

    @abc.abstractmethod
    def _get_test_sets(
        self, hide_targets=True, featurization_fn: Callable | None = None
    ) -> dict[str, Subset]:
        raise NotImplementedError

    @abc.abstractmethod
    def evaluate(
        self,
        y_pred: IncomingPredictionsType | None = None,
        y_prob: IncomingPredictionsType | None = None,
    ) -> BenchmarkResultsType:
        raise NotImplementedError

    def _get_subset(self, indices, hide_targets=True, featurization_fn=None) -> Subset:
        """Returns a [`Subset`][polaris.dataset.Subset] using the given indices. Used
        internally to construct the train and test sets."""
        return Subset(
            dataset=self.dataset,
            indices=indices,
            input_cols=self.input_cols,
            target_cols=self.target_cols,
            hide_targets=hide_targets,
            featurization_fn=featurization_fn,
        )

    def upload_to_hub(
        self,
        settings: PolarisHubSettings | None = None,
        cache_auth_token: bool = True,
        access: AccessType = "private",
        owner: HubOwner | str | None = None,
        parent_artifact_id: str | None = None,
        **kwargs: dict,
    ):
        """
        Very light, convenient wrapper around the
        [`PolarisHubClient.upload_benchmark`][polaris.hub.client.PolarisHubClient.upload_benchmark] method.
        """
        from polaris.hub.client import PolarisHubClient

        with PolarisHubClient(
            settings=settings,
            cache_auth_token=cache_auth_token,
            **kwargs,
        ) as client:
            return client.upload_benchmark(
                self, access=access, owner=owner, parent_artifact_id=parent_artifact_id
            )

    def to_json(self, destination: str) -> str:
        """Save the benchmark to a destination directory as a JSON file.

        Warning: Multiple files
            Perhaps unintuitive, this method creates multiple files in the destination directory as it also saves
            the dataset it is based on to the specified destination.

        Args:
            destination: The _directory_ to save the associated data to.

        Returns:
            The path to the JSON file.
        """
        dest = Path(destination)
        dest.mkdir(parents=True, exist_ok=True)

        data = self.model_dump()
        data["dataset"] = self.dataset.to_json(destination)

        path = dest / "benchmark.json"
        with fsspec.open(str(path), "w") as f:
            json.dump(data, f)

        return str(path)

    def _repr_dict_(self) -> dict:
        """Utility function for pretty-printing to the command line and jupyter notebooks"""
        repr_dict = self.model_dump()
        repr_dict["dataset_name"] = self.dataset.name
        return repr_dict

    def _repr_html_(self):
        """For pretty printing in Jupyter."""
        return dict2html(self.model_dump())

    def __repr__(self):
        return self.model_dump_json(indent=2)

    def __str__(self):
        return self.__repr__()


@deprecated(
    "Use BenchmarkV2Specification instead. If you're loading this dataset from the Polaris Hub, you can ignore this warning."
)
class BenchmarkV1Specification(
    SplitSpecificationV1Mixin, ChecksumMixin, BenchmarkSpecification[BenchmarkResultsV1]
):
    _version: ClassVar[Literal[1]] = 1

    dataset: DatasetV1 = Field(exclude=True)

    @field_validator("dataset", mode="before")
    @classmethod
    def _parse_dataset(cls, v: DatasetV1 | str | dict[str, Any]) -> DatasetV1:
        """
        Allows either passing a Dataset object or the kwargs to create one
        """
        match v:
            case dict():
                return DatasetV1(**v)
            case str():
                return DatasetV1.from_json(v)
            case DatasetV1():
                return v

    @model_validator(mode="after")
    def _infer_target_types(self) -> Self:
        """Try to automatically infer the target types if not already set"""

        for target in filter(lambda target: target not in self.target_types, self.target_cols):
            # Skip inferring the target type for pointer columns.
            # This would be complex to implement properly.
            # For these columns, dataset creators can still manually specify the target type.
            column_annotation = self.dataset.annotations.get(target)
            if column_annotation is not None and column_annotation.is_pointer:
                continue

            val = self.dataset.table.loc[:, target]

            # Non-numeric columns can be targets (e.g. prediction molecular reactions),
            # but in that case we currently don't infer the target type.
            if not np.issubdtype(val.dtype, np.number):
                continue

            # Remove the nans for multiple task dataset when the table is sparse
            target_type = type_of_target(val[~np.isnan(val)])
            match target_type:
                case "continuous":
                    self.target_types[target] = TargetType.REGRESSION
                case "binary" | "multiclass":
                    self.target_types[target] = TargetType.CLASSIFICATION

        return self

    @model_validator(mode="after")
    def _validate_split_in_dataset(self) -> Self:
        # All indices are valid given the dataset. We check the len of `self` here because a
        # competition entity includes both the dataset and benchmark in one artifact.
        max_i = len(self.dataset)
        if any(i < 0 or i >= max_i for i in chain(self.split[0], *self.split[1].values())):
            raise InvalidBenchmarkError("The predefined split contains invalid indices")

        return self

    @model_validator(mode="after")
    def _validate_cols_in_dataset(self) -> Self:
        """
        Verifies that all specified columns are present in the dataset.
        """
        columns = self.target_cols | self.input_cols
        dataset_columns = set(self.dataset.columns)
        if not columns.issubset(dataset_columns):
            raise InvalidBenchmarkError("Not all target or input columns were found in the dataset.")

        return self

    def _compute_checksum(self) -> str:
        """
        Computes a hash of the benchmark.

        This is meant to uniquely identify the benchmark and can be used to verify the version.
        """
        hash_fn = md5()
        hash_fn.update(self.dataset.md5sum.encode("utf-8"))
        for c in sorted(self.target_cols):
            hash_fn.update(c.encode("utf-8"))
        for c in sorted(self.input_cols):
            hash_fn.update(c.encode("utf-8"))
        for m in sorted(str(m) for m in self.metrics):
            hash_fn.update(m.encode("utf-8"))

        # Train set
        s = json.dumps(sorted(self.split[0]))
        hash_fn.update(s.encode("utf-8"))

        # Test sets
        for k in sorted(self.split[1].keys()):
            s = json.dumps(sorted(self.split[1][k]))
            hash_fn.update(k.encode("utf-8"))
            hash_fn.update(s.encode("utf-8"))

        checksum = hash_fn.hexdigest()
        return checksum

    def _get_test_sets(
        self, hide_targets=True, featurization_fn: Callable | None = None
    ) -> dict[str, Subset]:
        """
        Construct the test set(s), given the split in the benchmark specification. Used
        internally to construct the test set for client use and evaluation.
        """
        test_split = self.split[1]
        return {
            k: self._get_subset(v, hide_targets=hide_targets, featurization_fn=featurization_fn)
            for k, v in test_split.items()
        }

    def get_train_test_split(
        self, featurization_fn: Callable | None = None
    ) -> tuple[Subset, Subset | dict[str, Subset]]:
        """Construct the train and test sets, given the split in the benchmark specification.

        Returns [`Subset`][polaris.dataset.Subset] objects, which offer several ways of accessing the data
        and can thus easily serve as a basis to build framework-specific (e.g. PyTorch, Tensorflow)
        data-loaders on top of.

        Args:
            featurization_fn: A function to apply to the input data. If a multi-input benchmark, this function
                expects an input in the format specified by the `input_format` parameter.

        Returns:
            A tuple with the train `Subset` and test `Subset` objects.
                If there are multiple test sets, these are returned in a dictionary and each test set has
                an associated name. The targets of the test set can not be accessed.
        """
        train = self._get_subset(self.split[0], hide_targets=False, featurization_fn=featurization_fn)
        test = self._get_test_sets(hide_targets=True, featurization_fn=featurization_fn)

        # For improved UX, we return the object instead of the dictionary if there is only one test set.
        # Internally, however, assume that the test set is always a dictionary simplifies the code.
        if len(test) == 1:
            test = test["test"]
        return train, test

    @computed_field
    @property
    def n_classes(self) -> dict[str, int]:
        """
        The number of classes for each of the target columns.
        """
        return {
            target: self.dataset.table.loc[:, target].nunique()
            for target in self.target_cols
            if self.target_types.get(target) == TargetType.CLASSIFICATION
        }

    def evaluate(
        self,
        y_pred: IncomingPredictionsType | None = None,
        y_prob: IncomingPredictionsType | None = None,
    ) -> BenchmarkResultsV1:
        """Execute the evaluation protocol for the benchmark, given a set of predictions.

        info: What about `y_true`?
            Contrary to other frameworks that you might be familiar with, we opted for a signature that includes just
            the predictions. This reduces the chance of accidentally using the test targets during training.

        For this method, we make the following assumptions:

        1. There can be one or multiple test set(s);
        2. There can be one or multiple target(s);
        3. The metrics are _constant_ across test sets;
        4. The metrics are _constant_ across targets;
        5. There can be metrics which measure across tasks.

        Args:
            y_pred: The predictions for the test set, as NumPy arrays.
                If there are multiple targets, the predictions should be wrapped in a dictionary with the target labels as keys.
                If there are multiple test sets, the predictions should be further wrapped in a dictionary
                    with the test subset labels as keys.
            y_prob: The predicted probabilities for the test set, formatted similarly to predictions, based on the
                number of tasks and test sets.

        Returns:
            A `BenchmarkResultsV1` object. This object can be directly submitted to the Polaris Hub.

        Examples:
            1. For regression benchmarks:
                pred_scores = your_model.predict_score(molecules) # predict continuous score values
                benchmark.evaluate(y_pred=pred_scores)
            2. For classification benchmarks:
                - If `roc_auc` and `pr_auc` are in the metric list, both class probabilities and label predictions are required:
                    pred_probs = your_model.predict_proba(molecules) # predict probablities
                    pred_labels = your_model.predict_labels(molecules) # predict class labels
                    benchmark.evaluate(y_pred=pred_labels, y_prob=pred_probs)
                - Otherwise:
                    benchmark.evaluate(y_pred=pred_labels)
        """

        # Instead of having the user pass the ground truth, we extract it from the benchmark spec ourselves.
        y_true = self._get_test_sets(hide_targets=False)

        scores = evaluate_benchmark(
            target_cols=list(self.target_cols),
            test_set_labels=self.test_set_labels,
            test_set_sizes=self.test_set_sizes,
            metrics=self.metrics,
            y_true=y_true,
            y_pred=y_pred,
            y_prob=y_prob,
        )

        return BenchmarkResultsV1(results=scores, benchmark_artifact_id=self.artifact_id)

    def __eq__(self, other) -> bool:
        if not isinstance(other, BenchmarkSpecification):
            return False
        return self.md5sum == other.md5sum



import numpy as np
import pandas as pd
import pytest

from polaris.benchmark import BenchmarkV1Specification
from polaris.dataset import Dataset
from polaris.evaluate._metric import Metric


def test_absolute_average_fold_error():
    y_true = np.random.uniform(low=50, high=100, size=200)
    y_pred_1 = y_true + np.random.uniform(low=0, high=5, size=200)
    y_pred_2 = y_true + np.random.uniform(low=5, high=20, size=200)
    y_pred_3 = y_true - 10
    y_zero = np.zeros(shape=200)

    metric = Metric(label="absolute_average_fold_error")
    # Optimal value
    aafe_0 = metric.fn(y_true=y_true, y_pred=y_true)
    assert aafe_0 == 1

    # small fold change
    aafe_1 = metric.fn(y_true=y_true, y_pred=y_pred_1)
    assert aafe_1 > 1

    # larger fold change
    aafe_2 = metric.fn(y_true=y_true, y_pred=y_pred_2)
    assert aafe_2 > aafe_1

    # undershoot
    aafe_3 = metric.fn(y_true=y_true, y_pred=y_pred_3)
    assert aafe_3 < 1

    # y_true contains zeros
    with pytest.raises(ValueError):
        metric.fn(y_true=y_zero, y_pred=y_pred_3)


def test_grouped_metric():
    metric = Metric(label="accuracy", config={"group_by": "group"})

    table = pd.DataFrame({"group": ["a", "b", "b"], "y_true": [1, 1, 1]})
    dataset = Dataset(table=table)
    benchmark = BenchmarkV1Specification(
        dataset=dataset,
        metrics=[metric],
        main_metric=metric,
        target_cols=["y_true"],
        input_cols=["group"],
        split=([], [0, 1, 2]),
    )

    result = benchmark.evaluate([1, 0, 0])

    # The global accuracy is only 33%, but because we compute it per group and then average, it's 50%.
    assert result.results.Score.values[0] == 0.5


def test_metric_hash():
    metric_1 = Metric(label="accuracy")
    metric_2 = Metric(label="accuracy")
    metric_3 = Metric(label="mean_absolute_error")
    assert hash(metric_1) == hash(metric_2)
    assert hash(metric_1) != hash(metric_3)

    metric_4 = Metric(label="accuracy", config={"group_by": "group1"})
    assert hash(metric_4) != hash(metric_1)

    metric_5 = Metric(label="accuracy", config={"group_by": "group2"})
    assert hash(metric_4) != hash(metric_5)

    metric_6 = Metric(label="accuracy", config={"group_by": "group1"})
    assert hash(metric_4) == hash(metric_6)


def test_metric_name():
    metric = Metric(label="accuracy")
    assert metric.name == "accuracy"

    metric = Metric(label="accuracy", config={"group_by": "group"})
    assert metric.name == "accuracy_grouped"

    metric.custom_name = "custom_name"
    assert metric.name == "custom_name"



import json

import fsspec
from datamol.utils import fs

from polaris.benchmark import MultiTaskBenchmarkSpecification, SingleTaskBenchmarkSpecification
from polaris.benchmark._benchmark_v2 import BenchmarkV2Specification
from polaris.dataset import DatasetV1, create_dataset_from_file
from polaris.hub.client import PolarisHubClient
from polaris.utils.types import ChecksumStrategy
from polaris.model import Model


def load_dataset(path: str, verify_checksum: ChecksumStrategy = "verify_unless_zarr") -> DatasetV1:
    """
    Loads a Polaris dataset.

    In Polaris, a dataset is a tabular data structure that stores data-points in a row-wise manner.
    A dataset can have multiple modalities or targets, can be sparse
    and can be part of _one or multiple benchmarks_.

    The Polaris dataset can be loaded from the Hub or from a local or remote directory.

    - **Hub** (recommended): When loading the dataset from the Hub, you can simply
        provide the `owner/name` slug. This can be easily copied from the relevant dataset
        page on the Hub.
    - **Directory**: When loading the dataset from a directory, you should provide the path
        as returned by `dataset.to_json()`. The path can be local or remote.
    """

    extension = fs.get_extension(path)
    is_file = fs.is_file(path) or extension == "zarr"

    if not is_file:
        # Load from the Hub
        with PolarisHubClient() as client:
            return client.get_dataset(*path.split("/"), verify_checksum=verify_checksum)

    # Load from local file
    if extension == "json":
        dataset = DatasetV1.from_json(path)
    else:
        dataset = create_dataset_from_file(path)

    # Verify checksum if requested
    if dataset.should_verify_checksum(verify_checksum):
        dataset.verify_checksum()

    return dataset


def load_benchmark(path: str, verify_checksum: ChecksumStrategy = "verify_unless_zarr"):
    """
    Loads a Polaris benchmark.

    In Polaris, a benchmark wraps a dataset with additional metadata to specify the evaluation logic.

    The Polaris benchmark can be loaded from the Hub or from a local or remote directory.

    Note: Dataset is automatically loaded
        The dataset underlying the benchmark is automatically loaded when loading the benchmark.

    - **Hub** (recommended): When loading the benchmark from the Hub, you can simply
        provide the `owner/name` slug. This can be easily copied from the relevant benchmark
        page on the Hub.
    - **Directory**: When loading the benchmark from a directory, you should provide the path
        as returned by `benchmmark.to_json()`. The path can be local or remote.
    """
    is_file = fs.is_file(path) or fs.get_extension(path) == "zarr"

    if not is_file:
        # Load from the Hub
        with PolarisHubClient() as client:
            return client.get_benchmark(*path.split("/"), verify_checksum=verify_checksum)

    with fsspec.open(path, "r") as fd:
        data = json.load(fd)

    is_single_task = isinstance(data["target_cols"], str) or len(data["target_cols"]) == 1

    match data["version"]:
        case 1 if is_single_task:
            cls = SingleTaskBenchmarkSpecification
        case 1:
            cls = MultiTaskBenchmarkSpecification
        case 2:
            cls = BenchmarkV2Specification
        case _:
            raise ValueError(f"Unsupported benchmark version: {data['version']}")

    benchmark = cls.from_json(path)

    # Verify checksum if requested
    if benchmark.dataset.should_verify_checksum(verify_checksum):
        benchmark.verify_checksum()

    return benchmark


def load_competition(artifact_id: str):
    """
    Loads a Polaris competition.

    On Polaris, a competition represents a secure and fair benchmark. The target labels never exist
    on the client and all results are evaluated through Polaris' servers.

    Note: Dataset is automatically loaded
        The dataset underlying the competition is automatically loaded when loading the competition.

    """
    with PolarisHubClient() as client:
        return client.get_competition(artifact_id)


def load_model(artifact_id: str) -> Model:
    """
    Loads a Polaris model.

    On Polaris, a model centralizes all data about a method and can be attached to different results.
    """
    with PolarisHubClient() as client:
        return client.get_model(artifact_id)



import pytest

import polaris as po
from polaris.benchmark._base import BenchmarkSpecification
from polaris.dataset._base import BaseDataset
from polaris.hub.settings import PolarisHubSettings

settings = PolarisHubSettings()


@pytest.mark.skipif(
    settings.username is None or settings.password is None,
    reason="This test case requires headless authentication to be set up",
)
def test_load_dataset_flow():
    dataset = po.load_dataset("polaris/hello-world")
    assert isinstance(dataset, BaseDataset)


@pytest.mark.skipif(
    settings.username is None or settings.password is None,
    reason="This test case requires headless authentication to be set up",
)
def test_load_benchmark_flow():
    benchmark = po.load_benchmark("polaris/hello-world-benchmark")
    assert isinstance(benchmark, BenchmarkSpecification)



from typing import Collection, Sequence

from pydantic import (
    BaseModel,
    Field,
    ValidationInfo,
    computed_field,
    field_serializer,
    field_validator,
    model_validator,
)
from typing_extensions import Self

from polaris.evaluate import Metric
from polaris.utils.errors import InvalidBenchmarkError
from polaris.utils.types import ColumnName, TargetType, TaskType


class PredictiveTaskSpecificationMixin(BaseModel):
    """A mixin for predictive task benchmarks.

    Attributes:
        target_cols: The column(s) of the original dataset that should be used as the target.
        input_cols: The column(s) of the original dataset that should be used as input.
        metrics: The metrics to use for evaluating performance.
        main_metric: The main metric used to rank methods. If `None`, this defaults to the first of the `metrics` field.
        target_types: A dictionary that maps target columns to their type. If not specified, this is automatically inferred.
    """

    target_cols: set[ColumnName] = Field(min_length=1)
    input_cols: set[ColumnName] = Field(min_length=1)
    metrics: set[Metric] = Field(min_length=1)
    main_metric: Metric | str
    target_types: dict[ColumnName, TargetType] = Field(default_factory=dict, validate_default=True)

    @field_validator("target_cols", "input_cols", mode="before")
    @classmethod
    def _parse_cols(cls, v: str | Sequence[str], info: ValidationInfo) -> set[str]:
        """
        Normalize columns input values to a set.
        """
        if isinstance(v, str):
            v = {v}
        else:
            v = set(v)
        return v

    @field_validator("target_types", mode="before")
    @classmethod
    def _parse_target_types(
        cls, v: dict[ColumnName, TargetType | str | None]
    ) -> dict[ColumnName, TargetType]:
        """
        Converts the target types to TargetType enums if they are strings.
        """
        return {
            target: TargetType(val) if isinstance(val, str) else val
            for target, val in v.items()
            if val is not None
        }

    @field_validator("metrics", mode="before")
    @classmethod
    def _validate_metrics(cls, v: str | Metric | Collection[str | Metric]) -> set[Metric]:
        """
        Verifies all specified metrics are either a Metric object or a valid metric name.
        Also verifies there are no duplicate metrics.

        If there are multiple test sets, it is assumed the same metrics are used across test sets.
        """
        if isinstance(v, str):
            v = {"label": v}
        if not isinstance(v, Collection):
            v = [v]

        def _convert(m: str | dict | Metric) -> Metric:
            if isinstance(m, str):
                return Metric(label=m)
            if isinstance(m, dict):
                return Metric(**m)
            return m

        v = [_convert(m) for m in v]

        unique_metrics = set(v)

        if len(unique_metrics) != len(v):
            raise InvalidBenchmarkError("The benchmark specifies duplicate metrics.")

        unique_names = {m.name for m in unique_metrics}
        if len(unique_names) != len(unique_metrics):
            raise InvalidBenchmarkError(
                "The metrics of a benchmark need to have unique names. Specify a custom name with Metric(custom_name=...)"
            )

        return unique_metrics

    @model_validator(mode="after")
    def _validate_main_metric_is_in_metrics(self) -> Self:
        if isinstance(self.main_metric, str):
            for m in self.metrics:
                if m.name == self.main_metric:
                    self.main_metric = m
                    break
        if self.main_metric not in self.metrics:
            raise InvalidBenchmarkError("The main metric should be one of the specified metrics")
        return self

    @field_serializer("metrics")
    def _serialize_metrics(self, value: set[Metric]) -> list[Metric]:
        """
        Convert the set to a list. Since metrics are models and will be converted to dict,
        they will not be hashable members of a set.
        """
        return list(value)

    @model_validator(mode="after")
    def _validate_target_types(self) -> Self:
        """
        Verifies that all target types are for benchmark targets.
        """
        columns = set(self.target_types.keys())
        if not columns.issubset(self.target_cols):
            raise InvalidBenchmarkError(
                f"Not all specified target types were found in the target columns. {columns} - {self.target_cols}"
            )
        return self

    @field_serializer("main_metric")
    def _serialize_main_metric(value: Metric) -> str:
        """
        Convert the Metric to it's name
        """
        return value.name

    @field_serializer("target_types")
    def _serialize_target_types(self, target_types):
        """
        Convert from enum to string to make sure it's serializable
        """
        return {k: v.value for k, v in target_types.items()}

    @field_serializer("target_cols", "input_cols")
    def _serialize_columns(self, v: set[str]) -> list[str]:
        return list(v)

    @computed_field
    @property
    def task_type(self) -> str:
        """The high-level task type of the benchmark."""
        v = TaskType.MULTI_TASK if len(self.target_cols) > 1 else TaskType.SINGLE_TASK
        return v.value



import logging
from itertools import chain

from pydantic import BaseModel, computed_field, field_serializer, model_validator
from typing_extensions import Self

from polaris.utils.errors import InvalidBenchmarkError
from polaris.utils.misc import listit
from polaris.utils.types import SplitType

logger = logging.getLogger(__name__)


class SplitSpecificationV1Mixin(BaseModel):
    """
    Mixin class to add a split field to a benchmark. This is the V1 implementation.

    The split is defined as a (train, test) tuple, where train is a list of indices and
    test is a dictionary that maps test set names to lists of indices.

    Warning: Scalability
        The simple list-based representation we use for the split in this first implementation doesn't scale well.
        We therefore worked on a V2 implementation that uses roaring bitmaps.
        See [`SplitSpecificationV2Mixin`][`polaris.experimental._split_v2.SplitSpecificationV2Mixin`] for more details.

    Attributes:
        split: The predefined train-test split to use for evaluation.
    """

    split: SplitType

    @model_validator(mode="after")
    def _validate_split(self) -> Self:
        """
        Verifies that:
          1) There are no empty test partitions
          2) There is no overlap between the train and test set
          3) There is no duplicate indices in any of the sets
        """

        if not isinstance(self.split[1], dict):
            self.split = self.split[0], {"test": self.split[1]}
        split = self.split

        # Train partition can be empty (zero-shot)
        # Test partitions cannot be empty
        if any(len(v) == 0 for v in split[1].values()):
            raise InvalidBenchmarkError("The predefined split contains empty test partitions")

        train_idx_list = split[0]
        full_test_idx_list = list(chain.from_iterable(split[1].values()))

        if len(train_idx_list) == 0:
            logger.info(
                "This benchmark only specifies a test set. It will return an empty train set in `get_train_test_split()`"
            )

        train_idx_set = set(train_idx_list)
        full_test_idx_set = set(full_test_idx_list)

        # The train and test indices do not overlap
        if len(train_idx_set & full_test_idx_set) > 0:
            raise InvalidBenchmarkError("The predefined split specifies overlapping train and test sets")

        # Check for duplicate indices within the train set
        if len(train_idx_set) != len(train_idx_list):
            raise InvalidBenchmarkError("The training set contains duplicate indices")

        # Check for duplicate indices within a given test set. Because a user can specify
        # multiple test sets for a given benchmark and it is acceptable for indices to be shared
        # across test sets, we check for duplicates in each test set independently.
        for test_set_name, test_set_idx_list in split[1].items():
            if len(test_set_idx_list) != len(set(test_set_idx_list)):
                raise InvalidBenchmarkError(
                    f'Test set with name "{test_set_name}" contains duplicate indices'
                )

        return self

    @field_serializer("split")
    def _serialize_split(self, v: SplitType):
        """Convert any tuple to list to make sure it's serializable"""
        return listit(v)

    @computed_field
    @property
    def test_set_sizes(self) -> dict[str, int]:
        """The sizes of the test sets."""
        return {k: len(v) for k, v in self.split[1].items()}

    @computed_field
    @property
    def n_test_sets(self) -> int:
        """The number of test sets"""
        return len(self.split[1])

    @computed_field
    @property
    def n_train_datapoints(self) -> int:
        """The size of the train set."""
        return len(self.split[0])

    @computed_field
    @property
    def test_set_labels(self) -> list[str]:
        """The labels of the test sets."""
        return sorted(list(self.split[1].keys()))

    @computed_field
    @property
    def n_test_datapoints(self) -> dict[str, int]:
        """The size of (each of) the test set(s)."""
        if self.n_test_sets == 1:
            return {"test": len(self.split[1]["test"])}
        else:
            return {k: len(v) for k, v in self.split[1].items()}
